import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import processing.serial.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class NLEDConfigurationSoftwarev2b extends PApplet {

/*
 MIT License
 
 Copyright (c) 2018 Northern Lights Electronic Design, LLC
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 
 Original Author: Jeffrey Nygaard
 Company: Northern Lights Electronic Design, LLC
 Contact: JNygaard@NLEDShop.com
 Date Updated: September 5, 2018
 Software Version:  2b
 Webpage: www.NLEDShop.com/nledconfig
 Written in Processing v3.3.7  - www.Processing.org
 
 //============================================================================================================
 
 Supported Devices: 
 
 NLED Pixel Controller Mini - version 4 firmware
 NLED Pixel Controller Micro
 NLED Pixel Controller Ion
 NLED Pixel Controller Electron
 
 NLED 16-Bit RGBW Lamp Controller
 NLED 30 Channel High Current LED Controller
 NLED Quasar
 
Legacy Supported Devices: Note, if not working correctly use NLED Aurora Control Version 1
NLED Four Channel Controller Mini
NLED 4 Channel DMX Wash Controller
*/


/*
 To Do:
 
 
 Notes:
 
 - Mini 9600 baud
 - 4 Chan DMX at 19200
 - Octo at 19200
 
 Future Revisions:
 - TCP stack
 - merge with bootloader maybe...
 - Add presets like 30 chan DMX, or infinity mirrors  
 - Have it read back configurations and test them against what was sent
 */

// IMPORT Libraries


//============ Software Constants ================

final int cModueOffsetX = 20;
final int cModueOffsetY = 150;

//============ Application Constants ================


final String cDeviceSelectionList[] = {"Select Device From List(Default)", "Pixel Controller Mini(v4)", "Pixel Controller Micro(all)" };
final int cDeviceSelectionListIDNum[] = {0, 60, 61}; //stores aurora id numbers for the above strings

final String[] cPixelChipsetStrings = {"DEFAULT", "WS2801", "WS2811", "WS2812", "WS2812B", "LPD1886", "LPD6803", "LPD8806", "TM1803", "TM1814", "APA102", "APA104", "SK6812", "DMX-512", "APA106"};
final String[] cPixelChipsetColorID = {"varies / unknown", "varies / R>G>B", "varies", "G>R>B", "G>R>B", "TEST", "TEST", "TEST", "varies / G>B>R", "TEST", "B>G>R", "TEST", "G>R>B>W", "R>G>B", "R>G>B"};
final String cModeString[] = {"Stall", "Fade", "Gradient", "Instant", "Technical", "POV", "File Play", "Linked", "DMX", "Serial", "USB"};
final String cPixelIDStr[] = {"Single", "RGB", "RGBW"};
final String cNoneStr[] = {"None"};
final String cColorSwapStr[] = {"None", "B>R>G", "G>B>R", "R>B>G", "B>G>R", "G>R>B"};
final String cGlobalColSwapStr[] = {"R>G>B(>W)", "B>R>G", "G>B>R", "R>B>G", "B>G>R", "G>R>B", "G>R>B>W"};
final String cDDBaudstring[] = {"9600", "19200", "38400", "57600", "115200", "230400", "250000", "460800", "500000", "1Mbit"}; //fix these, there is no 250k without custom FTDI driver, but maybe make one for 500k and 1Mb and 250k
final String cDDLEDMstring[] = {"Both", "Stand-Alone", "RX", "Disabled"};
final String cDDPOVModeString[] = {"Normal Mode", "Mode 1mS", "Mode 2mS", "Mode 3mS", "Mode 4mS", "Mode 5mS", "Mode 6mS", "Mode 8mS", "Mode 16mS"};
final int cDDPOVModeValues[] = {0, 1, 2, 3, 4, 5, 6, 8, 16};
final String cRevisionIDstr[] = {"a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z"};
final String cDDAutoDetectModesT1[] = {"None/Disabled", "Auto-Detect DMX", "Auto-Detect Serial"};
final String cDDAutoDetectModesT2[] = {"None/Disabled", "Auto-Detect DMX"};
final String cDDAutoDetectModesT3[] = {"None/Disabled", "Auto-Detect"};
final String cDDMasterModes8Bit[] = {"None/Slave", "Master Full Pkts", "Master Partial Pkts"};
final String cDDMasterModesMultiBit[] = {"None/Slave", "Master Full Pkts 8-Bit", "Master Full Pkts 16-Bit", "Master Partial Pkts 8-Bit", "Master Partial Pkts 16-Bit"};
final String cDDDisplayModes[] = {"Always On", "Countdown", "Dim", "Countdown+Dim"};
final String cDDShapeOptions[] = {"Round", "Square", "Spiral", "Line", "Pixel Map", "Manual"};
final String cDDAccMeterShutOff[] = {"Disabled", "5 Sec.", "10 Sec.", "20 Sec.", "30 Sec", "40 Sec.", "60 Sec.", "80 Sec."};
final String cDDAccMeterShutOffMode[] = {"Blank", "Idle Sequence"};
final int cAccMeterShutOffVals[] = {0, 16, 32, 64, 96, 128, 192, 250};
final String cDDPowerTimeOut[] = {"Disabled", "5 Min.", "10 Min.", "15 Min."};
final String cDDAccMeterTapMode[] = {"Disabled", "Sensitive", "Normal", "Strong"};
final String cDDPWMFreqA[] = {"31 KHz", "7.8 KHz", "1.95 KHz", "488 Hz"};
final String cDDPWMProfileA[] = {"16-Bit, 244Hz", "12-Bit, 3.9KHz", "12-Bit, 478Hz", "8-Bit, 62.5Khz", "8-Bit, 976Hz"};
final String cDDIRRemote[] = {"Disabled", "Enabled, 19200 baud", "Enabled 250k baud"};
final String cDDStandAloneA[] = {"Hold", "Blank", "Fade Out", "Stand-Alone"};
final String cDDDMXRelstring[] = {"Disabled", "Timeout & Continue", "Timeout & Play Idle"};

final String cDDAutoReleaseLegacy[] = {"Disabled", "Enabled"};
final String cDDReceptionModeLegacy[] = {"DMX-512", "Serial"};

final String cDDColorSwapPixMicro[] = {"None/RGB", "G>R>B", "G>R>B>W", "B>G>R", "B>R>G"};
final String cDDBaudstringMiniV4[] = {"9600", "19200", "57600", "115200", "230400", "500000", "1Mbit"}; 

final String cDDPortBaudstring[] = {"USB/None", "9600", "19200", "38400", "57600", "115200", "230400", "250000", "460800", "500000", "1000000"};

final String cDDMicroDMXDecoderMode[] = {"Disabled", "Full Packets", "Force Mode", "Force Mode, Full Packets"};

final String cConfirmBoxText[] = {"null", "Uploading Configurations will overwrite the current settings. Continue?"};


//============ Variables ================

int mouseXS, mouseYS; //scaled mouse values for resizing
float SF = 1; //scale factor for GUI resizing

int OverlayMenuID = 0;

int modulePosX = 90;
int modulePosY = 380;
int moduleHoldX = 0;
int moduleHoldY = 0;
int moduleHoldSz = 0;
int[] PixelChipsetID  = new int[2];

boolean GlobalDDOpen = false;
boolean GlobalDragging = false;
boolean TextFieldActive = false;
boolean NumberInputFieldActive = false;

String GlobalLabelStore = "";  //used for text fields

String DisplayMessageStr = "Starting"; 

String[] DevStrArray = new String[16];
int ConfigFlags = 0;  
int holdHardwareID = 0;

int RXByte1 = -1;    // Incoming serial data
int RXByte2 = 0;
int RXByte3 = 0;
int RXByte4 = 0;
int RXByte5 = 0;

int SendCounter = 0;
int SendCounterB = 0;
int USBpacketCount = 0;

byte[] USBPacket = new byte[64];
int PacketPointer = 0;

boolean devConnected = false;
boolean RecievedOpenAck = false;
boolean SentCmdRequest = false;
boolean SentOpenRequest = false;
boolean SentConfigRequest = false;
boolean LiveControlEnabled = false;
boolean UploadInProgress = false;
boolean WaitForAckFlag = false;  
boolean TerminateUpload = false;

boolean ConfigUploadSent = false; //new

int ExpectedRecieved = 0;
int ReceiveCounter = 0;
int cmdFlags = 0;
int cmdByte = 4; //open connection default
int cmdData[] = {0, 0, 0, 0};

int RecievedDeviceConfigsMSB = 0;
int RecievedDeviceConfigsLSB = 0;


int CommunicationMode = 0; //0 = serial port, 1 = TCP Client
int ProgramBaudRate = 19200; 

int CMDTimeOutVal = 3000; //time in miliseconds default
boolean CmdIssued = false;

String portName = "None";


String ConfirmBoxCallBack = "";
int ConfirmBoxIDNum = 0;

//============ Aurora Command ID#s ================

int cmdOpen = 4;
int cmdUploadConfigurations = 101;
int cmdRequestConfigurations = 120;

//============ General Objects ================

GUIObj gui; //gui object, holds colors and such
SoftwareObj software;
DeviceObj device;

DropDown DropDownPointer;
SliderBar SliderPointer;
TextField textFieldPtr;
guiNumberInputField numberInputFieldPtr;

OverlayMenu[] guiOverlayMenus;

Serial serialPort;

SpecHWModules[] InputModule;  //Configuration Modules

//============ GUI Objects ================

Button menuCloseButton, ConfirmButtonYes, ConfirmButtonNo;

Button comAutoScanButton, comConnectButton, comUploadConfigurations, comDefaultConfigurations;

DropDown comSerialPortDD, comManualDeviceSelectDD, comSerialBaudRateDD;

//============ Module GUI Objects ================

DropDown DMXModesDD, PixelChipsetDD, AutoDetectDD, MasterModeDD, HardwareColSwapDD;
DropDown BaudRateDD, LEDIndicatorModeADD, LEDIndicatorModeBDD;
DropDown StandAloneDD, DisplayModeDD, DisplayModeBDD, ShapeOptionsDD;
DropDown AccMeterAutoDetectDD, AccMeterAutoDetectModeDD, AccMeterTapDD, PowerDownTimeOutDD;
DropDown PWMProfileADD, IRRemoteModeDD, StandAloneADD, PWMFreqADD, DMXAutoReleaseDD;

//added
DropDown BaudRateMiniV4DD, HardwareColSwapMicroDD, DMXDecoderModeDD;
DropDown AutoReleaseLegacyDD, ReceptionModeLegacyDD;

TextField[] TextFields;

Button[] HardwareButtons;
CheckBox[] HardwareCheckBox;

//================================== ArrayList for GUI MouseOver Elements ====================================================

//ArrayList<SourceContentTile> SourceContentTileList; //makes list to make it easier to mouseover
//SourceContentTile ContentTilePtr;

ArrayList<Button> ButtonList; //makes list to make it easier to mouseover
Button PointerButton;

ArrayList<TextField> TextFieldList;
TextField TextFieldListPointer;

ArrayList<CheckBox> CheckBoxList;
CheckBox CheckBoxListPointer;

ArrayList<DropDown> DropDownList; //makes list to make it easier to mouseover
int DropDownMouseOverID = 0;

PFont font;

//========================= END OBJECT DECLARTION ==============================

public void setup() 
{

  // Initial window size
   
  surface.setResizable(true);   // Needed for resizing the window to the sender size
  surface.setLocation(100, 100);
  colorMode(RGB);
  frameRate(30);

  font = createFont("Arial", 48);
  textFont(font);

  PImage titlebaricon = loadImage("favicon.gif");
  surface.setIcon(titlebaricon);

  //==================== Init ArrayList for objects ======================

  ButtonList = new ArrayList<Button>();  
  DropDownList = new ArrayList<DropDown>();  
  TextFieldList = new ArrayList<TextField>();  
  CheckBoxList = new ArrayList<CheckBox>();  

  //====================== Init General Objects ==========================

  gui = new GUIObj(); //init graphic user interface from XML file
  software = new SoftwareObj();
  device = new DeviceObj("No Device");

  BuildCOMDropDown();//comSerialPortDD = new DropDown(serialPort.list(), serialPort.list().length, 0, 10, 10, 130, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  comManualDeviceSelectDD = new DropDown(cDeviceSelectionList, cDeviceSelectionList.length, 0, 400, 25, 300, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "ManualDeviceDDCallBack");
  comSerialBaudRateDD = new DropDown(cDDPortBaudstring, cDDPortBaudstring.length, 0, 240, 45, 120, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");

  ConfirmButtonYes = new Button("Yes", 325, 325, 60, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, false, true);
  ConfirmButtonNo = new Button("No", 415, 325, 60, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, false, true);
  comAutoScanButton = new Button("Auto-Scan Ports", 240, 10, 120, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, false, true);
  comConnectButton = new Button("Connect", 140, 10, 90, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, false, true);
  comUploadConfigurations  = new Button("Upload Configurations", 20, 110, 160, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, false, true);
  comDefaultConfigurations  = new Button("Set To Default", 625, 110, 140, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, false, true);

  comDefaultConfigurations.status = 1; //disable for now, no function
  comAutoScanButton.status  = 1;//disable for now, no function

  //====================== Start GUI objects Definition ==========================

  AccMeterAutoDetectDD = new DropDown(cDDAccMeterShutOff, cDDAccMeterShutOff.length, 0, 420, 180, 100, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  AccMeterAutoDetectModeDD = new DropDown(cDDAccMeterShutOffMode, cDDAccMeterShutOffMode.length, 0, 420, 180, 100, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  AccMeterTapDD  = new DropDown(cDDAccMeterTapMode, cDDAccMeterTapMode.length, 0, 420, 180, 100, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  LEDIndicatorModeADD = new DropDown(cDDLEDMstring, cDDLEDMstring.length, 0, 420, 180, 100, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  LEDIndicatorModeBDD = new DropDown(cDDLEDMstring, cDDLEDMstring.length, 0, 420, 180, 100, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  DMXModesDD = new DropDown(cNoneStr, 1, 0, 85, 280, 180, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  BaudRateDD = new DropDown(cDDBaudstring, cDDBaudstring.length, 0, 260, 215, 100, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  DisplayModeDD = new DropDown(cDDDisplayModes, cDDDisplayModes.length, 0, 415, 240, 120, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");    
  DisplayModeBDD = new DropDown(cDDDisplayModes, cDDDisplayModes.length, 0, 415, 240, 120, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");    
  MasterModeDD = new DropDown(cDDMasterModes8Bit, cDDMasterModes8Bit.length, 0, 90, 340, 140, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");  
  AutoDetectDD = new DropDown(cDDAutoDetectModesT1, cDDAutoDetectModesT1.length, 0, 405, 300, 120, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");  
  HardwareColSwapDD = new DropDown(cGlobalColSwapStr, cGlobalColSwapStr.length, 0, 480, 470, 100, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  PixelChipsetDD = new DropDown(cNoneStr, cNoneStr.length, 0, 350, 470, 140, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack"); 
  ShapeOptionsDD = new DropDown(cDDShapeOptions, cDDShapeOptions.length, 0, 290, 425, 100, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  PowerDownTimeOutDD = new DropDown(cDDPowerTimeOut, cDDPowerTimeOut.length, 0, 405, 300, 120, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");  
  PWMProfileADD = new DropDown(cDDPWMProfileA, cDDPWMProfileA.length, 0, 405, 300, 140, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");  
  IRRemoteModeDD = new DropDown(cDDIRRemote, cDDIRRemote.length, 0, 405, 300, 140, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");  
  PWMFreqADD = new DropDown(cDDPWMFreqA, cDDPWMFreqA.length, 0, 5, 160, 100, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");   
  StandAloneADD = new DropDown(cDDStandAloneA, cDDStandAloneA.length, 0, 5, 160, 110, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");  
  //StandAloneDD = new DropDown(cColorSwapStr, cColorSwapStr.length, 0, 215, 60, 120, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  DMXAutoReleaseDD = new DropDown(cDDDMXRelstring, cDDDMXRelstring.length, 0, 10, 190, 130, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false,"genericDDCallBack");
  
  //Added
  BaudRateMiniV4DD = new DropDown(cDDBaudstringMiniV4, cDDBaudstringMiniV4.length, 0, 260, 215, 100, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  HardwareColSwapMicroDD = new DropDown(cDDColorSwapPixMicro, cDDColorSwapPixMicro.length, 0, 480, 470, 100, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  DMXDecoderModeDD = new DropDown(cDDMicroDMXDecoderMode, cDDMicroDMXDecoderMode.length, 0, 480, 470, 140, 20, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");

  AutoReleaseLegacyDD = new DropDown(cDDAutoReleaseLegacy, cDDAutoReleaseLegacy.length, 0, 10, 190, 100, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false,"genericDDCallBack");
  ReceptionModeLegacyDD =  new DropDown(cDDReceptionModeLegacy, cDDReceptionModeLegacy.length, 0, 10, 190, 100, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false,"genericDDCallBack");

  //=====================================================================================================================================

  TextFields = new TextField[15];

  //TextFields[0] = new TextField("Enter#", 35, 510, 50, 25, color(255), color(200, 40, 40), 1, 0, 255, true, false, "txtHandleSlideValue"); //Slide Value
  //don't have to do this TextFields[0].Status = 1; //start greyed out
  TextFields[1] = new TextField("Enter#", 225, 470, 50, 20, color(255), color(200, 40, 40), 1, 0, 100, true, false, "txtHandleGenericNumber"); //whats this for now? 
  TextFields[2] = new TextField("1", 120, 195, 80, 20, color(255), color(200, 40, 40), 1, 1, 512, true, false, "txtHandleGenericNumber"); //DMX Adr Config

  TextFields[3] = new TextField("5", 120, 195, 80, 20, color(255), color(200, 40, 40), 1, 5, 32, true, false, "txtHandleGenericNumber");  // CONVERTED - End-Of-Frame Timer
  TextFields[4] = new TextField("170", 120, 195, 80, 20, color(255), color(200, 40, 40), 1, 1, 170, true, false, "txtHandleGenericNumber");  // CONVERTED - Pixel Amount

  TextFields[13] = new TextField("1", 20, 265, 25, 25, color(255), color(200, 40, 40), 1, 1, 32, true, false, "txtHandleGenericNumber");  //Stand-Alone Module ID field

  textFieldPtr =  TextFields[1]; //default it so it doesn't NullPointerException

  //=====================================================================================================================================

  HardwareButtons = new Button[6];
  HardwareButtons[0] = new Button("Enabled", 125, 470, 80, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, true, false, true);
  HardwareButtons[1] = new Button("Enabled", 300, 470, 80, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, true, false, true);
  HardwareButtons[2] = new Button("Enabled", 300, 470, 80, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, true, false, true);
  HardwareButtons[3] = new Button("Enabled", 255, 330, 80, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, true, false, true);
  HardwareButtons[4] = new Button("Enabled", 300, 470, 80, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, true, false, true);  

  HardwareCheckBox = new CheckBox[4];
  HardwareCheckBox[0] = new CheckBox(255, 285, 20, 20, color(255), color(0), color(0), false);
  HardwareCheckBox[1] = new CheckBox(255, 285, 20, 20, color(255), color(0), color(0), false);

  //added
  HardwareCheckBox[2] = new CheckBox(255, 285, 20, 20, color(255), color(0), color(0), true); //enable activity LED
  HardwareCheckBox[3] = new CheckBox(255, 285, 20, 20, color(255), color(0), color(0), false); //enable serial color swap - pixmicro

  //====================== Start Module Definition ==========================

  InputModule = new SpecHWModules[34];
  InputModule[0] = new SpecHWModules(0, "Serial RS-485 Enable:\n(Disable for TTL Serial)", 0, 0, 140, 90, "fRS485Enable", "BYTE", 0);
  InputModule[1] = new SpecHWModules(1, "Second Activity\nLED Mode:", 0, 0, 130, 90, "fLEDModeMSBb,fLEDModeLSBb", "BYTE", 0);
  InputModule[2] = new SpecHWModules(2, "Select Correct Color Order For Chipset:", 0, 0, 120, 90, "CONFIG", "bGlobalColorShift", 210);
  InputModule[3] = new SpecHWModules(3, "Select The Pixel Chipset:\n\n\nRecommended Color: varies / R>G>B", 0, 0, 160, 90, "CONFIG", "bPixelChipset", 250);
  InputModule[4] = new SpecHWModules(4, "Enable Serial Reception Instead of DMX reception. Auto-Detect is disabled during serial reception.", 0, 0, 200, 90, "CONFIG", "BYTE", 0);
  InputModule[5] = new SpecHWModules(5, "DMX Address:\n(Enter Number)\n\n\n(1-512)", 0, 0, 120, 90, "CONFIG", "bDMXAdrMSB,bDMXAdrLSB", 0);
  InputModule[6] = new SpecHWModules(6, "DMX Reception\nMode - ID Number: "+DMXModesDD.selStr, 0, 0, 190, 90, "CONFIG", "bDMXModeID", 190); 
  InputModule[7] = new SpecHWModules(7, "DMX Master Mode:\n(Not all listed options are valid)", 0, 0, 160, 90, "fMasterEnable,fMasterFullPkts,fMaster16Bit", "BYTE", 190); 
  InputModule[8] = new SpecHWModules(8, "Serial Baud Rate:\nID Number: "+BaudRateDD.selStr, 0, 0, 120, 90, "f8or16Bit", "bBaudRateID", 200); 
  InputModule[9] = new SpecHWModules(9, "Activity\nLED Mode:", 0, 0, 120, 90, "fLEDModeMSBa,fLEDModeLSBa", "BYTE", 0); 
  InputModule[10] = new SpecHWModules(10, "External\nLED Display Mode:", 0, 0, 140, 90, "fDisModeMSBa,fDisModeLSBa", "BYTE", 190); 
  InputModule[11] = new SpecHWModules(11, "Auto-Detect Data Control Reception:", 0, 0, 140, 90, "fAutoDetectID,fAutoDetect", "BYTE", 180); 
  InputModule[12] = new SpecHWModules(12, "Enable Gamma\nCorrection:\n(See Connection tab)", 0, 0, 140, 90, "fGammaCorrectionEnabled", "CONFIG", 0); 
  InputModule[13] = new SpecHWModules(13, "Accelerometer Movement Detection:", 0, 0, 140, 90, "fAccMeterAutoShutOff,fAccMeterAutoShutOffMode", "bAccmeterTimeoutVal", 210); 
  InputModule[14] = new SpecHWModules(14, "Enable Accelerometer Speed Adjustment", 0, 0, 140, 90, "fAccMeterMovement", "BYTE", 0); 
  InputModule[15] = new SpecHWModules(15, "Enable Accelerometer Double Tap Button Press", 0, 0, 140, 90, "fAccMeterDoubleTap", "bAccmeterDoubleTapVal", 200); 
  InputModule[16] = new SpecHWModules(16, "Power Down\nTime Out", 0, 0, 140, 90, "fTimeOutEnabledMSB,fTimeOutEnabledLSB", "BYTE", 0); 
  InputModule[17] = new SpecHWModules(17, "Pixel Data Packet Cloning:\n(0 is Disabled)", 0, 0, 120, 90, "DELETE", "DELETE", 0); 
  InputModule[18] = new SpecHWModules(18, "Sequence Fade Transition:\n(0 is Disabled)", 0, 0, 120, 90, "DELETE", "DELETE", 0); 
  InputModule[19] = new SpecHWModules(19, "Dual Communication Mode (Serial\n/WiFi/Bluetooth)", 0, 0, 120, 90, "fDualModeCom", "BYTE", 0); 
  InputModule[20] = new SpecHWModules(20, "PWM Frequency:", 0, 0, 120, 90, "BIT", "bPWMModeLegacyA", 180); 
  InputModule[21] = new SpecHWModules(21, "Stand-Alone Mode\n& Signal Loss Mode:", 0, 0, 160, 90, "CONFIG", "bStandaloneLegacyA", 160);
  InputModule[22] = new SpecHWModules(22, "PWM Profile\nResolution & Frequency:", 0, 0, 160, 90, "CONFIG", "bPWMMode", 160);
  InputModule[23] = new SpecHWModules(23, "Enable I.R. Remote Control:\n(Addon Card)", 0, 0, 160, 90, "fIRCardEnable,fIRCardEnableSpeed", "BYTE", 160);

  //added for configuration software  
  InputModule[24] = new SpecHWModules(24, "DMX Pixel Amount:\n(Enter Number)\n\n\n(1-170)", 0, 0, 120, 90, "CONFIG", "bUserPixelAmountMSB,bUserPixelAmountLSB", 0); //textfield
  InputModule[25] = new SpecHWModules(25, "Enable Activity LED:", 0, 0, 120, 90, "fActivityLEDEnable", "BYTE", 0); //checkbox
  InputModule[26] = new SpecHWModules(26, "DMX Decoder Mode:", 0, 0, 160, 90, "fDMXDecoderModeFull,fDMXDecoderModeEnable", "BYTE", 0); //Drop down
  InputModule[27] = new SpecHWModules(27, "Enable Serial Color Swap:", 0, 0, 120, 90, "fEnableSerialColorSwap", "BYTE", 0); //checkbox
  InputModule[28] = new SpecHWModules(28, "Select Color Order For Chipset:", 0, 0, 120, 90, "fColorOrderGRB,fColorOrderGRBW,fColorOrderBGR,fColorOrderBRG", "BYTE", 210); //drop down
  InputModule[29] = new SpecHWModules(29, "End-Of-Frame Timer:\n(in miliseconds):", 0, 0, 120, 90, "CONFIG", "bUserTimerValue", 0); //textfield
  InputModule[30] = new SpecHWModules(30, "Serial Baud Rate:\nID Number: "+BaudRateDD.selStr, 0, 0, 120, 90, "CONFIG", "bBaudRateIDMiniV4", 200); //Drop down

//out of order vs NLED Aurora
  InputModule[31] = new SpecHWModules(31,"DMX Timeout\n Release:",0,0, 150, 90,"fDMXAutoRelease,fDMXReleaseIdle","BYTE",100);
  
  //new legacy modules
  InputModule[32] = new SpecHWModules(32,"Signal Loss Detection\n(Runs Signal Loss Action):",0,0, 140, 90,"AutoDetectFlag","BYTE",100);
  InputModule[33] = new SpecHWModules(33,"Reception Mode:",0,0, 120, 90,"SerialOrDMX","BYTE",100);  
  
  //set default message
  DisplayMessageStr = "Software Ready"; //rather than call show notificaiton...
  
  /*
 //Start Force a Device to load at launch
 device.HardwareID = 43;
 device.FirmwareVersion = 1;
 device.FirmwareRevision = 4;
  try {
    LoadDeviceFile();
  }
  catch(Exception e) { 
    println("CAUGHT");
    ShowNotification(16);
  }
  //End Development Force
  */
  
} //end setup()

//======================================================================================================

public void BuildCOMDropDown()
{
  //store and reselect value, say COM 6, and COM4 is added teh selStr is wrong, find the proper value again
  //String holdCOM = 

  if (devConnected == false) ShowNotification(2);

  comSerialPortDD = new DropDown(serialPort.list(), serialPort.list().length, 0, 10, 10, 120, 25, gui.buttonColor, gui.buttonHighlightColor, gui.textColor, false, "genericDDCallBack");
  
  //shorten serial name string if too long
      for (int i=0; i != Serial.list().length;i++)
    {
      if (comSerialPortDD.labels[i].length() > 10) 
      {
      comSerialPortDD.labels[i] = comSerialPortDD.labels[i].substring(comSerialPortDD.labels[i].length()-10, comSerialPortDD.labels[i].length());
      }
    }
}

//======================================================================================================

//===================================================================================================  

public void SendConfigurations()
{
  println("SendConfigurations()"); 

  String[] ModuleString = new String[16];  
  int[] TransmitByteArray = new int[16];  
  DevStrArray = split(device.ConfigFlagsStr, ',');
  //Now single delimenated string is an array, the pointer is the bit number in ConfigFlags  

  //Module0 with matching string ex:   "fDisModeMSBa,fDisModeLSBa"
  //Module1 with matching string ex:   "fAutoDetectID,fAutoDetect"
  //Loop goes through enabled modules, checks the string to the DevStr array  
  //  If a module string matches the list build ConfigFlags

  ConfigFlags = 0; //reset flags here.....

  for(int y = 0; y != InputModule.length;y++)
  {  
    if(InputModule[y].Enabled == true)
    {
      ModuleString = split(InputModule[y].ModuleFlagStr, ',');
      
      for(int q = 0; q != ModuleString.length; q++)
      {  
        for(int i = 0; i != DevStrArray.length; i++)
        {  
          if(DevStrArray[i].equals(ModuleString[q]) == true) //if string matches
          {
            //println("String Matches "+ModuleString[q]+" on Bit ID# "+i+"   StrID: "+q);
            
            if(InputModule[y].GetBitValues(q) == 1) bitSet(i); //this pretty much does it gotta setup modules
          }
        } //end i for()
      } //end q for()  
    } //end if
  } //end y for()

  println("ConfigFlags: "+binary(ConfigFlags,16));
  
  for(int i = 0; i != TransmitByteArray.length;i++) TransmitByteArray[i] = 0; //clear array first
  TransmitByteArray[0] = (ConfigFlags >> 8); //MSB first
  TransmitByteArray[1] = ConfigFlags & 0xFF;  //LSB second
  ///NOW RUN CONFIGURATION BYTES  

  DevStrArray = split(device.ConfigBytesStr, ',');  //load device config byte string
  
  for(int y = 0; y != InputModule.length;y++)
  {    
    if(InputModule[y].Enabled == true)
    {
      ModuleString = split(InputModule[y].ModuleByteStr, ',');

      for(int q = 0; q != ModuleString.length; q++)
      {  
        for(int i = 0; i != DevStrArray.length; i++)
        {  
          if(DevStrArray[i].equals(ModuleString[q]) == true) //if string matches
          {
            //println("Set Byte - String Matches "+ModuleString[q]+" on BYTE# "+(i+2));
            TransmitByteArray[i+2] = InputModule[y].GetByteValues(q);
          }
        } //end i for()
      } //end q for()  
    } //end if
  } //end y for()
  
  printArray(TransmitByteArray);


if(device.BasicVersion == 0)
{
  //Now Send TransmitByteArray
  for(int i = 0; i != (DevStrArray.length+2); i++)  SendSingleDataByte(TransmitByteArray[i]); //+2 for 2x config bytes
}
else if(device.BasicVersion == 3)
{
   println("Sending legacy configurations, rearranging"); 
    //Puts DMX address first, dumps the config MSB, 
  TransmitByteArray[15] = TransmitByteArray[1]; //store first byte in temp - store CONFIG LSB
  TransmitByteArray[0] = TransmitByteArray[2]; //move DMXMSB
  TransmitByteArray[1] = TransmitByteArray[3]; //move DMXLSB
  TransmitByteArray[2] = TransmitByteArray[15]; //restore saved config flags lsb
  TransmitByteArray[3] = TransmitByteArray[4];
  TransmitByteArray[4] = TransmitByteArray[5];
  TransmitByteArray[5] = TransmitByteArray[6];
  TransmitByteArray[6] = TransmitByteArray[7];  
  TransmitByteArray[7] = 0;
  //printArray(TransmitByteArray);
  //Now Send TransmitByteArray
  for(int i = 0; i != 7; i++)  SendSingleDataByte(TransmitByteArray[i]); //magic number
}
  println("SendConfigurations() finished");  
} //end SendConfigurations()  


//=============================================================================================================================================  

 //mini order
 //unlock sequence -> PixelChipsetID(0) -> AdrLSB -> AdrMSB - > UserPixLo -> UserPixHi -> BaudRateID -> custom SPBRG value -> User PR2 Value -> ConfigFlags(8)
 
 // micro
 //unlock sequence -> Chipset ID# -> User Pixel Amount -> ConfigFlags
  
public void SendConfigurationsNonAurora()
{
  println("SendConfigurationsNonAurora()"); 

  String[] ModuleString = new String[16];  
  int[] TransmitByteArray = new int[16];  
  DevStrArray = split(device.ConfigFlagsStr, ',');
  //Now single delimenated string is an array, the pointer is the bit number in ConfigFlags  

  //Module0 with matching string ex:   "fDisModeMSBa,fDisModeLSBa"
  //Module1 with matching string ex:   "fAutoDetectID,fAutoDetect"
  //Loop goes through enabled modules, checks the string to the DevStr array  
  //  If a module string matches the list build ConfigFlags

  ConfigFlags = 0; //reset flags here.....

  for(int y = 0; y != InputModule.length;y++)
  {  
    if(InputModule[y].Enabled == true)
    {
      ModuleString = split(InputModule[y].ModuleFlagStr, ',');
      
      for(int q = 0; q != ModuleString.length; q++)
      {  
        for(int i = 0; i != DevStrArray.length; i++)
        {  
          if(DevStrArray[i].equals(ModuleString[q]) == true) //if string matches
          {
            //println("String Matches "+ModuleString[q]+" on Bit ID# "+i+"   StrID: "+q);
            
            if(InputModule[y].GetBitValues(q) == 1) bitSet(i); //this pretty much does it gotta setup modules
          }
        } //end i for()
      } //end q for()  
    } //end if
  } //end y for()

  println("ConfigFlags: "+binary(ConfigFlags,16));
  
  for(int i = 0; i != TransmitByteArray.length;i++) TransmitByteArray[i] = 0; //clear array first
  
  
//  TransmitByteArray[0] = (ConfigFlags >> 8); //MSB first
 // TransmitByteArray[1] = ConfigFlags & 0xFF;  //LSB second
  ///NOW RUN CONFIGURATION BYTES  

  DevStrArray = split(device.ConfigBytesStr, ',');  //load device config byte string

  
  for(int y = 0; y != InputModule.length;y++)
  {    
    if(InputModule[y].Enabled == true)
    {
      ModuleString = split(InputModule[y].ModuleByteStr, ',');

      for(int q = 0; q != ModuleString.length; q++)
      {  
        for(int i = 0; i != DevStrArray.length; i++)
        {  
          if(DevStrArray[i].equals(ModuleString[q]) == true) //if string matches
          {
            //println("Set Byte - String Matches "+ModuleString[q]+" on BYTE# "+(i+2));
            TransmitByteArray[i] = InputModule[y].GetByteValues(q);
          }     
        } //end i for()
      } //end q for()  
    } //end if
  } //end y for()
  
  //now add config flags
  
  TransmitByteArray[DevStrArray.length] = ConfigFlags & 0xFF;  //LSB second
  
  println("Configs placed in ["+(DevStrArray.length)+"]    Sending "+(DevStrArray.length+1)+" bytes");
  printArray(TransmitByteArray);

  //Send Unlock sequence - 123 -> 253 -> 82 -> 7
  SendSingleDataByte(123);
  SendSingleDataByte(253);
  SendSingleDataByte(82);
  SendSingleDataByte(7);

  //Now Send TransmitByteArray
  for(int i = 0; i != (DevStrArray.length+1); i++)  SendSingleDataByte(TransmitByteArray[i]);

  println("SendConfigurationsNonAurora() finished");  
 // ShowNotification(28);
} //end SendConfigurations()  

//==================================================================================================================================================
//===================================================================================================

public void OpenCOMPort(String NameID)
{
  println("OpenCOMPort() with "+NameID+" at baud: "+ProgramBaudRate);

  if (NameID.equals("TCP/IP"))
  {
    thread("OpenTCP");
    return;
  }

  try {
    serialPort.stop();
    serialPort.dispose();
  }
  catch(Exception e) {
  }

  int x = 0; //so can be used in catch

  try 
  {
    for (x =0; x != Serial.list().length; x++) 
    {
      if (NameID.equals(Serial.list()[x]))
      {  
        portName = Serial.list()[x];      
        serialPort = new Serial(this, portName, ProgramBaudRate);

        ShowNotification(3);
        println("Connected to "+portName);

        //Decide here what to do about micro and mini which don't use the aurora protocol

        if (comManualDeviceSelectDD.selStr == 0) OpenConnection(); //Aurora device selected, attempt connection command
        else
        {
          //either a mini or micro, handle
          devConnected = true;    


          device.HardwareID = cDeviceSelectionListIDNum[comManualDeviceSelectDD.selStr];
          //
          // if(comManualDeviceSelectDD.selStr == 1) device.HardwareID = 60;
          //   else if(comManualDeviceSelectDD.selStr == 2) device.HardwareID = 61;       

          device.HardwareVersion = 1;
          if (device.HardwareID == 60) device.FirmwareVersion = 4;
          else device.FirmwareVersion = 1;
          device.FirmwareRevision = 0;   

          LoadDeviceFile();
          ShowNotification(7);
        }

        break;
      } else
      {
        //println(Serial.list()[x]+" is not a Match for COM port in software.ini");
      }
    }
  }
  catch(Exception e)
  {
    println("Could Not Open COM Port "+Serial.list()[x]);
    devConnected = false;    
    portName = "None";
    ShowNotification(22);
  }
} //end func()

//===================================================================================================

public void CloseCOMPort()
{
  println("CloseCOMPort() - Closing Port");    
  devConnected = false;    

  try   
  {  
    //close COM port
    serialPort.stop();
    serialPort.dispose();
    devConnected = false;
  }
  catch(Exception e)
  {
    println("No Port To Close");
  }  

  portName = "None";
  println("COM Port Closed");
}

//===================================================================================================

public void CloseCommunicationPort()
{
  if (CommunicationMode == 0) CloseCOMPort();  
  //  else if(CommunicationMode == 1) CloseTCP();
} //end CloseCommunicationPort()


//===================================================================================================

//separate so one can pass characters, for exchange, other for data/numbers
public void SendSingleDataChar(char passedValue)
{
  try {
    if (CommunicationMode == 0) serialPort.write(passedValue);
    //  else if(CommunicationMode == 1) TCPClient.write(passedValue);
  }
  catch(Exception e) {
  }
}

public void SendSingleDataByte(int passedValue)
{
  println("Sent 1 byte: "+passedValue);

  try {
    if (CommunicationMode == 0) serialPort.write(passedValue);
    // else if(CommunicationMode == 1) TCPClient.write(passedValue);
  }
  catch(Exception e) {
  }
}


public void SendAckedCommand() //threaded
{
  println("Cmd Request Acknowledged, Sending - CmdByte: "+cmdByte+" : "+cmdData[0]+" : "+cmdData[1]+" : "+cmdData[2]+" : "+cmdData[3]);
  RecievedOpenAck = false;
  SendSingleDataByte(cmdByte);
  SendSingleDataByte(cmdData[0]);  
  SendSingleDataByte(cmdData[1]);   
  SendSingleDataByte(cmdData[2]);  
  SendSingleDataByte(cmdData[3]); 

  CmdIssued = true; //set flag to wait for authentication
}

//===================================================================================================  

public void serialEvent(Serial serialPort) 
{
  RXByte5 = RXByte4;
  RXByte4 = RXByte3;
  RXByte3 = RXByte2;
  RXByte2 = RXByte1;
  RXByte1 = serialPort.read();
  DataRecieved(); //run common function
}  //end serialEvent()

//===================================================================================================  

public void DataRecieved()  
{
  println("Received: "+RXByte1+"  Char: "+PApplet.parseChar(RXByte1));
  //START TIMER METHOD
  /*
int tempTime = (millis() - HoldMiliSec);
   println("Received: "+RXByte1+"/"+char(RXByte1)+" Time: "+tempTime);
   HoldMiliSec=millis();  
   */
  //END TIMER METHOD

  if (SentCmdRequest == true && RXByte2 == 'a' && RXByte1 == '9')
  {
    RecievedOpenAck = true;
    SentCmdRequest = false;
    EnableTransmissionTimeOut();
    //myPort.write("nled99"); 
    SendSingleDataChar('n'); 
    SendSingleDataChar('l');  
    SendSingleDataChar('e');
    SendSingleDataChar('d');  
    SendSingleDataChar('9');   
    SendSingleDataChar('9');

    println("Sent Acknowledge");
  } 
  else if (RecievedOpenAck == true  && RXByte2 == 'f' && RXByte1 == '0')
  {
    thread("SendAckedCommand");
  } 
  else if (SentConfigRequest == true)
  {
    switch(ReceiveCounter)
    {
    case 0: //MSB
      RecievedDeviceConfigsMSB = RXByte1;
      break;      
    case 1: //LSB
      RecievedDeviceConfigsLSB = RXByte1;    
      SentConfigRequest = false;
      println("Requested and Recieved Configurations "+binary(RecievedDeviceConfigsMSB, 8)+" "+binary(RecievedDeviceConfigsLSB, 8));
      break;
    } //end switch
    ReceiveCounter++;
  } 
  else if (SentOpenRequest == true)
  {
    println("Counter: "+ReceiveCounter);
    switch(ReceiveCounter)
    {
    case 0: //hardware ID
      holdHardwareID = device.HardwareID; //holds current or old HardwareID

      device.HardwareID = RXByte1;
      break;      
    case 1: //hardware version ID, not really used
      device.HardwareVersion = RXByte1;
      break;
    case 2: //Firmware Version
      device.FirmwareVersion = RXByte1;
      break;
    case 3: //Firmware Revision
      device.FirmwareRevision = RXByte1;
      break;
    case 4:  //Bootloader Target HardwareID

      break;      
    case 5:  //Bootloader Version Number
      device.BootloaderVersion = RXByte1;
      break;      
    case 6:  //UserID Number Configuration
      device.UserConfiguredIDNum = RXByte1;

      //All Device Connection bytes received, now load device file and continue
      SentOpenRequest = false;
      devConnected = true;
      //SaveSoftwareConfigFile();  //save COMPort Name

      try {
        LoadDeviceFile();
        //String temp = cRevisionIDstr[device.FirmwareRevision]; //to test and catch if valid firmware revision number
      }
      catch(Exception e) { 
        ShowNotification(16);
      }

      //if(device.FirmwareVersion > 100) ConfirmBoxIDNum = 11; //unverified firmware
      //redraw();
      break;
    }  
    ReceiveCounter++;
  } 
  else if (UploadInProgress == true || ConfigUploadSent == true) //added the cmdUploadConfigurations check so it would ack config upload
  {
    if (RXByte4 == 'z' && RXByte3 == 'A' && RXByte2 == 'c' && RXByte1 == 'k')  
    {
      WaitForAckFlag = true;
      println("ACK RECEIVED FOR PACKET# "+USBpacketCount);
      //also send CRC byte and packet# to maintain continuity
      if (ConfigUploadSent == true) ShowNotification(5);
      ConfigUploadSent = false;
    }
    if (RXByte2 == 'C' && RXByte1 == 'S')  
    {
      //legacy configurations upload acknowledge
      WaitForAckFlag = true;
      println("Legacy Device Acknowledged Confiugrations upload");
      //also send CRC byte and packet# to maintain continuity
      if (ConfigUploadSent == true) ShowNotification(5);
      ConfigUploadSent = false;
    }   
  } 
  else if (CmdIssued == true)
  {
    if (RXByte5 == 'c' && RXByte4 == 'm' && RXByte3 == 'd' && RXByte2 == 0)  
    {
      CmdIssued = false;  
      println("COMMAND #"+RXByte1+" AUTHENTICATED");
      //run common init
      SendCounter = 0;
      SendCounterB = 0;
      USBpacketCount = 0; //reset counter  
      PacketPointer = 0; //would also get set to 0, since it always sends 64 bytes then clears it.

      //this executes the actual command software side once ACKed
      switch(cmdByte)
      {     
      case 4: //Request Connection
        SentOpenRequest = true;
        ReceiveCounter = 0;
        break;
      case 101:  //Upload Configurations
        UploadInProgress = true;  
        ConfigUploadSent = true;
        SendConfigurations();  
        UploadInProgress = false; 
        break;
      case 120: //Request Configurations
        SentConfigRequest = true;
        ReceiveCounter = 0;
        break;
      }  //end switch
    } //end byte check
    //else if(devConnected == false && cmdByte == 4)
    else if (RXByte4 > 0 && RXByte3 == 1 && RXByte2 == 1 && RXByte1 < 26)  
    {
      //Detects if version 1, rather than acking the cmd
      //hardwareID,hardwarev, firmwareV, firmwareRev

      println("Version 1 controller detected");
      // ShowNotification(30);

      println(RXByte1+"   "+RXByte2+"   "+RXByte3+"   "+RXByte4+"   "+RXByte5);

      //RXByte4 = ID numnber
      //RXByte3 = hardware v
      //RXByte2 = firmware v
      //RXByte1 = firmware rev

      device.HardwareID = RXByte4;
      device.HardwareVersion = RXByte3;
      device.FirmwareVersion = RXByte2;
      device.FirmwareRevision = RXByte1;    

      SentOpenRequest = false;
      devConnected = true;

      try {
        LoadDeviceFile();
        //String temp = cRevisionIDstr[device.FirmwareRevision]; //to test and catch if valid firmware revision number
      }
      catch(Exception e) { 
        ShowNotification(16);
      }

      //CloseCommunicationPort(); //can't from here
      //CloseCOMFlag = true;
    }
  } //end CmdIssued if()
} //end func

//=====================================================================================================

public void RequestCommand()
{
  TerminateUpload = false; //reset flag
  WaitForAckFlag = false; 

  try {
    // SendSingleDataByte("NLED11"); 
    SendSingleDataChar('N'); 
    SendSingleDataChar('L');  
    SendSingleDataChar('E');   
    SendSingleDataChar('D');  
    SendSingleDataChar('1');   
    SendSingleDataChar('1');
    SentCmdRequest = true; 
    EnableTransmissionTimeOut();
  }
  catch(Exception e)
  {
    println("Port Not Open, Or Unable to Send"); 
    SentCmdRequest = false;

    CloseCommunicationPort();
    ShowNotification(8);
  }
} //end func()

//===================================================================================================  

public void OpenConnection()
{
  println("Attempting Connection");

  cmdByte = cmdOpen; //connect to device
  cmdData[0] = 0;
  cmdData[1] = 0; 
  cmdData[2] = 0;  
  cmdData[3] = 0;  
  devConnected = false;

  RequestCommand();
}//end func()

//===================================================================================================  

public void RequestDeviceConfigurations()
{
  println("RequestDeviceConfigurations()");

  cmdByte = cmdRequestConfigurations; 
  cmdData[0] = 0;
  cmdData[1] = 0; 
  cmdData[2] = 0;  
  cmdData[3] = 0;  

  RequestCommand();
}//end func()

//=====================================================================================================

public void RequestConfigurationUploadType2()
{
  println("RequestConfigurationUploadType2() for Basic Version Type 2");

  cmdByte = cmdUploadConfigurations; //Upload Configurations
  cmdData[0] = 0; //MSB

  switch(PowerDownTimeOutDD.selStr) //Power Down Timer Mode
  {
  case 0: //disabled
    cmdData[0] = 0x00;
    break;
  case 1: //5 min
    cmdData[0] = 0x01;
    break;
  case 2: //10 min
    cmdData[0] =  0x02;
    break;
  case 3: //15 min
    cmdData[0] = 0x03;
    break;
  }  

  //RequestCmdTimeOut = CMDTimeOutVal;  

  cmdData[1] = 0; //LSB
  cmdData[2] = 0;  
  cmdData[3] = 0;  
  RequestCommand();
}

//===================================================================================================  

public void RequestConfigurationUploadLegacy()
{
  println("RequestConfigurationUpload()");

  cmdByte = 1; //Upload Configurations use old command ID value
  cmdData[0] = 0; //MSB
  cmdData[1] = 0; //LSB
  cmdData[2] = 0;  
  cmdData[3] = 0;  
  RequestCommand();
  delay(100); //wait for device to setup, no command confirm
  UploadInProgress = true;  
  ConfigUploadSent = true;
  SendConfigurations();  
  UploadInProgress = false; 
} //end RequestConfigurationUpload

//===================================================================================================  

public void RequestConfigurationUpload()
{
  println("RequestConfigurationUpload()");

  cmdByte = cmdUploadConfigurations; //Upload Configurations
  cmdData[0] = 0; //MSB
  cmdData[1] = 0; //LSB
  cmdData[2] = 0;  
  cmdData[3] = 0;  
  RequestCommand();
} //end RequestConfigurationUpload
//===================================================================================================  

//===================================================================================================  

public void draw() {
  //update in case of slider or other gui features
  mouseXS = round((float)mouseX / SF);
  mouseYS = round((float)mouseY / SF);

  background(gui.windowBackground);

  strokeWeight(2);
  stroke(gui.windowStroke);
  line(0, 100, 800, 100);

  fill(gui.textColor);
  textSize(14);
  textAlign(LEFT);


  if (devConnected == true) 
  {
    text("Device Connected", 10, 75);
    text("Serial Port: "+portName+" at "+ProgramBaudRate+" baud", 10, 60);
  } else  
  {
    text("Device Not Connected", 10, 75);
    text("Serial Port: "+portName, 10, 60);
  }

  text("Status: "+DisplayMessageStr, 10, 90);
  textSize(16);
  text("NLED Configuration Software v.2b", 400, 20);

  if (devConnected == true)
  {
    text("Device: "+device.Name, 400, 40);
    text("Firmware Version: "+(device.FirmwareVersion)+(PApplet.parseChar(device.FirmwareRevision+97)), 400, 60);
    //if(comManualDeviceSelectDD.selStr > 0) text("This device does not respond, and may not actually be connected",400,80);
  }


  //======= Handle GUI elements ======================
  //could be done in a better way but this is quicker and less debugging

  //if(portName.equals("None")) comAutoScanButton.status  = 0;
  //else comAutoScanButton.status  = 1;

  if (devConnected == false) 
  {
    text("Some Devices Require Manual Selection", 400, 70);
    comUploadConfigurations.status = 1;
    //comDefaultConfigurations.status  = 1;
    comConnectButton.Label = "Connect";
  } else
  {
    comUploadConfigurations.status = 0;
    //comDefaultConfigurations.status  = 0;
    comConnectButton.Label = "Disconnect";
  }
  //=================== END =========================

  for (int i =0; i != InputModule.length; i++) InputModule[i].display();

  comAutoScanButton.display();
  comConnectButton.display();

  comUploadConfigurations.display();
  comDefaultConfigurations.display();

  comSerialPortDD.display();

  if (devConnected == false) 
  {
    comManualDeviceSelectDD.display();
    comSerialBaudRateDD.display();
  }


  if (serialPort.list().length != comSerialPortDD.numStrs) BuildCOMDropDown(); //updates drop down

  if (GlobalDDOpen == true) DropDownPointer.display(); //if a drop down is open, display() it again to ensure it overlays all other drawing

  //---------------------------------------------------------------------------------------------------------------------

  if (ConfirmBoxIDNum > 0)
  {
    fill(gui.windowBackground, 128);
    rect(0, 0, width, height);

    fill(gui.windowStroke);
    stroke(255, 0, 0);
    strokeWeight(3);
    rect((800/2)-120, (600/2)-60, 240, 120, 20);
    fill(0);
    textSize(14);
    textAlign(CENTER);
    text(cConfirmBoxText[ConfirmBoxIDNum], (800/2)-110, (600/2)-50, 220, 100);
    textAlign(LEFT);

    ConfirmButtonYes.display();
    ConfirmButtonNo.display();
  }  

  //---------------------------------------------------------------------------------------------------------------------
} //end draw()

//===================================================================================================  

public void SetConfirmBox(int passedVal)
{
  ConfirmBoxIDNum = passedVal;
}

//============================================================================================

public void ShowNotification(int passedVal)
{
  println("ShowNotification() with "+passedVal);

  switch(passedVal)
  {
  case 0: //null

    break;
  case 1:
    DisplayMessageStr = "Connection and Definition Loading Successful.";
    break;
  case 2:
    DisplayMessageStr = "Available Serial Ports Have Updated.";
    break;    
  case 3:
    DisplayMessageStr = "Port Opened Successfully. Attempting Connection, please wait.";
    break; 
  case 4:
    DisplayMessageStr = "Configurations uploaded, waiting for acknowledge.";
    break;    
  case 5:
    DisplayMessageStr = "Configurations Successfully Uploaded";
    break;    
  case 6:
    DisplayMessageStr = "Command timed out. Device did not respond. Disconnecting from port.";
    break; 
  case 7:
    DisplayMessageStr = "This device does not respond, and may not actually be connected. See activity LED and datasheet for details.";
    break; 
   case 16:
    DisplayMessageStr = "Could not load device file, firmware may be outdated or not compatible. See www.NLEDshop.com/deviceupdates";
    break;   
  case 22:
    DisplayMessageStr = "Port Can Not Be Opened.";
    break;
    
  }//end switch
}

//============================================================================================

public void TransmissionTimerThread()
{
  int myStart = millis();

  while ((millis() - myStart) < CMDTimeOutVal)
  {
    delay(1);  //Required or would only work a few times, then stops working
    if (SentCmdRequest == false && RecievedOpenAck == false) { 
      return;
    } //force out by mouse click after min time
  }

  CloseCommunicationPort();

  ShowNotification(6); //display notification

  WaitForAckFlag = false;
  UploadInProgress = false;
  RecievedOpenAck = false;  
  SentCmdRequest = false;  

  println("Command Timed Out "+(millis() - myStart));
  redraw();
} //end NotificationTimerThread()

public void EnableTransmissionTimeOut()
{
  println("EnableTransmissionTimeOut()");
  redraw();
  thread("TransmissionTimerThread");
} //end func

//================================= START CONFIGURATION MODULES GUI ELEMENTS=================================================================

public boolean CheckModuleElements()
{
  //==================================================================================================
  if (InputModule[2].Enabled == true && HardwareColSwapDD.over()) return true; 
  //==================================================================================================
  if (InputModule[3].Enabled == true && PixelChipsetDD.over()) { FuncRecommendColorShift(); return true; }
  //==================================================================================================  
  if (InputModule[9].Enabled == true && LEDIndicatorModeADD.over()) return true;
  //==================================================================================================  
  if (InputModule[1].Enabled == true && LEDIndicatorModeBDD.over()) return true;
  //==================================================================================================  
  if (InputModule[10].Enabled == true && DisplayModeDD.over()) return true;  
  //==================================================================================================  
  if (InputModule[7].Enabled == true && MasterModeDD.over()) return true;  
  //==================================================================================================  
  if (InputModule[11].Enabled == true && AutoDetectDD.over()) return true;
  //==================================================================================================  
  if (InputModule[22].Enabled == true && PWMProfileADD.over()) return true;
  //==================================================================================================  
  if (InputModule[23].Enabled == true && IRRemoteModeDD.over()) return true;
  //==================================================================================================  
  if (InputModule[8].Enabled == true && BaudRateDD.over()) return true;
  //==================================================================================================  
  if (InputModule[6].Enabled == true && DMXModesDD.over()) return true;
  //==================================================================================================  
  if (InputModule[14].Enabled == true && AccMeterAutoDetectDD.over()) return true;
  //==================================================================================================  
  if (InputModule[14].Enabled == true && AccMeterAutoDetectModeDD.over()) return true;
  //==================================================================================================  
  if ( InputModule[15].Enabled == true && AccMeterTapDD.over()) return true;
  //=================================================================================================  
  if (InputModule[16].Enabled == true && PowerDownTimeOutDD.over()) return true;
  //==================================================================================================    
  if (InputModule[20].Enabled == true && PWMFreqADD.over()) return true;
  //==================================================================================================    
  if (InputModule[21].Enabled == true && StandAloneADD.over()) return true;   
  //==================================================================================================    
  if (InputModule[31].Enabled == true && DMXAutoReleaseDD.over()) return true;
  //==================================================================================================    


  if (InputModule[32].Enabled == true && AutoReleaseLegacyDD.over()) return true;
  //==================================================================================================    
  if (InputModule[33].Enabled == true && ReceptionModeLegacyDD.over()) return true;
  //==================================================================================================    



  if (HardwareCheckBox[0].over() && InputModule[8].Enabled == true) return true;//16-bit serial checkbox, not always enabled via Status
  if (HardwareCheckBox[1].over() && InputModule[19].Enabled == true) return true; //Dual Communication Mode

  //==================================================================================================

  if (HardwareButtons[0].over() && InputModule[0].Enabled == true) return true;//Enable/Disable RS485

  if (HardwareButtons[1].over() && InputModule[1].Enabled == true) return true;  //16 or 8 bit DMX Transmission

  if (HardwareButtons[2].over() && InputModule[4].Enabled == true) return true;

  if (HardwareButtons[3].over() && InputModule[12].Enabled == true) return true;  //gamma   

  if (HardwareButtons[4].over() && InputModule[14].Enabled == true) return true;    

  //==================================================================================================  

  if (TextFields[2].over() && InputModule[5].Enabled == true) { 
    TextFields[2].SelectField(); 
    return true;
  }//DMX Address
  if (TextFields[13].over() && InputModule[21].Enabled == true) { 
    TextFields[13].SelectField(); 
    return true;
  }//StandAlone Field A    

  //==================================================================================================  

  //added for configuration software  

  if (HardwareCheckBox[2].over() && InputModule[25].Enabled == true) return true;  //enable activity LED

  if (HardwareCheckBox[3].over() && InputModule[27].Enabled == true) return true;  //Enable Serial Color Swap:

  if (TextFields[4].over() && InputModule[24].Enabled == true) { 
    TextFields[4].SelectField(); 
    return true;
  }//DMX pixel amount

  if (TextFields[3].over() && InputModule[29].Enabled == true) { 
    TextFields[3].SelectField(); 
    return true;
  }//End of Frame Timer
  
    //==================================================================================================  
  if (InputModule[26].Enabled == true && DMXDecoderModeDD.over()) return true;
      //==================================================================================================  
  if (InputModule[28].Enabled == true && HardwareColSwapMicroDD.over()) return true;
        //==================================================================================================  
  if (InputModule[30].Enabled == true && BaudRateMiniV4DD.over()) return true;
  
  //==================================================================================================  

  return false;
} //end func


//===================================================================================================

public void FuncRecommendColorShift()
{  
  //String[] PixelChipsetStrings = {"WS2801", "WS2811", "WS2812", "WS2812B", "LPD1886", "LPD6803", "LPD8806", "TM1803", "TM1814", "APA102", "APA104", "SK6812"};

  try
  {
    for (int i=0; i<=PixelChipsetDD.numStrs+1; i++)
    {
      if (PixelChipsetDD.labels[PixelChipsetDD.selStr].indexOf(cPixelChipsetStrings[i])>=0)
      {
        InputModule[3].Label = "Select The Pixel Chipset:\n\n\nRecommended Color: "+cPixelChipsetColorID[i];
        return; //return if something is found
      }
    } //end for()
  }
  catch(Exception e)
  {
    //nothing
  }
  //otherwise uses default
  InputModule[3].Label = "Select The Pixel Chipset:\n\n\nRecommended Color: Unknown / Test";
} //end FuncRecommendColorShift

//===================================================================================================

public void LoadDeviceFile()
{
  //Called when device responds to a Connection Request
  //Loads a TXT file with the device specific defines and variables, loads into the device object
 
  if(device.HardwareID == 0)
  {
    //Null
    device.Name = "No Device";
    return;
  }
  
  String BuildString = "devices"+File.separator+""+device.HardwareID+"-nled-contr-revs.txt";   //loads file with current firmware rev info

  String[] lines = loadStrings(BuildString); 
  String[] WorkString = new String[30]; //may need to up this  
  
  BuildString = "";  
  
  for(int i = 0; i != lines.length;i++)
  {
    WorkString = split(lines[i], '\t');
    println(PApplet.parseInt(WorkString[0])+" : "+PApplet.parseInt(WorkString[1]));    
    
    if(PApplet.parseInt(WorkString[0]) == device.FirmwareVersion)//version
    {

      if(PApplet.parseInt(WorkString[1]) < device.FirmwareRevision && i == 0)//version
      {
        println("Revison higher than local "+device.FirmwareRevision);
        BuildString = "devices"+File.separator+""+device.HardwareID+"-"+PApplet.parseInt(WorkString[0])+"-"+PApplet.parseInt(WorkString[1])+"-nled-contr.txt";
        break;
      } //end [1] if
      

      if(PApplet.parseInt(WorkString[1]) == device.FirmwareRevision)//version
      {
        println("REVISION FOUND with "+i);
        BuildString = "devices"+File.separator+""+device.HardwareID+"-"+PApplet.parseInt(WorkString[0])+"-"+PApplet.parseInt(WorkString[1])+"-nled-contr.txt";
        break;
      } //end [1] if

    } //end [0] if
  }//end for()

  //detects if a revID is lower than what is found, use the highest
  if(BuildString.equals(""))
  {
    println("Device File For V. & Rev. Not Found - Using Most Recent Version");  
    lines = loadStrings("devices"+File.separator+""+device.HardwareID+"-nled-contr-revs.txt"); 
    WorkString = split(lines[0], '\t');
    BuildString = "devices"+File.separator+""+device.HardwareID+"-"+PApplet.parseInt(WorkString[0])+"-"+PApplet.parseInt(WorkString[1])+"-nled-contr.txt";  
  }

  println("LoadDeviceFile() - "+BuildString);  

  lines = loadStrings(BuildString); 
  int LinePointer = 1;
  
  //WorkString[0] is HardwareID, already checked....
  WorkString = split(lines[LinePointer++], '\t');
  device.Name = WorkString[1]; //String Name
  
  WorkString = split(lines[LinePointer++], '\t');
  device.WebpageURL = WorkString[1];    //String URL


  WorkString = split(lines[LinePointer++], '\t');
  if(device.FirmwareVersion < PApplet.parseInt(WorkString[1]))
  {
    println("FIRMWARE OUTDATED "+PApplet.parseInt(WorkString[1])+" vs "+device.FirmwareVersion);
    if(devConnected == true) //if no device is connected don't show the message
    {  
    //  ConfirmBoxIDNum = 8;
    }
  }
  
  WorkString = split(lines[LinePointer++], '\t');
  if(device.FirmwareRevision < PApplet.parseInt(WorkString[1]))
  {
    println("FIRMWARE REVISION OUTDATED "+PApplet.parseInt(WorkString[1])+" vs "+device.FirmwareRevision);
    if(devConnected == true) //if no device is connected don't show the message
    {
     // ConfirmBoxIDNum = 8;
    }
  }  

  //device.HardwareVersion = int(WorkString[1]);
  LinePointer++;  //for hardwareversion which is not read

  WorkString = split(lines[LinePointer++], '\t');
  device.Channels = PApplet.parseInt(WorkString[1]);   
  
  WorkString = split(lines[LinePointer++], '\t');
  device.LockChannelAmt = PApplet.parseBoolean(WorkString[1]);   
  
  WorkString = split(lines[LinePointer++], '\t');
  device.DataSpace = PApplet.parseInt(WorkString[1]);   
  
  WorkString = split(lines[LinePointer++], '\t');
  device.IndexBlockSize = PApplet.parseInt(WorkString[1]);
  try {   
    device.HPVBlockSize = PApplet.parseInt(WorkString[2]);   
    device.SeqMemoryBlockSize = PApplet.parseInt(WorkString[3]);   
  }
  catch(Exception e) 
  {
    println("Error loading block sizes, using single value");
    device.CurrentBlockSize  = device.IndexBlockSize;
    device.HPVBlockSize = device.CurrentBlockSize;   
    device.SeqMemoryBlockSize = device.CurrentBlockSize;    
  }
  
  WorkString = split(lines[LinePointer++], '\t');
  device.EraseBlockSize = PApplet.parseInt(WorkString[1]);  
  
  WorkString = split(lines[LinePointer++], '\t');
  device.HWPVSpace = PApplet.parseInt(WorkString[1]);    

  WorkString = split(lines[LinePointer++], '\t');
  device.MaxIndexedSequences = PApplet.parseInt(WorkString[1]);  
  
  WorkString = split(lines[LinePointer++], '\t');
  device.IndexMemoryModel = PApplet.parseInt(WorkString[1]);      
  
  WorkString = split(lines[LinePointer++], '\t');
  device.FillUpperData = PApplet.parseBoolean(WorkString[1]); 

  WorkString = split(lines[LinePointer++], '\t');
  device.Bit16Processor = PApplet.parseBoolean(WorkString[1]);    

  WorkString = split(lines[LinePointer++], '\t');
  device.Bit16Mode = PApplet.parseBoolean(WorkString[1]);    

  WorkString = split(lines[LinePointer++], '\t');
  device.BasicVersion = PApplet.parseInt(WorkString[1]);      
  
  WorkString = split(lines[LinePointer++], '\t');
  device.LinkedSupport = PApplet.parseBoolean(WorkString[1]);   
  
  WorkString = split(lines[LinePointer++], '\t');
  device.GammaCorrection = PApplet.parseInt(WorkString[1]);  //0=none, 1=8-bit gamma table, 2=16-bit calculation

  WorkString = split(lines[LinePointer++], '\t');
  device.AccelerometerMode = PApplet.parseInt(WorkString[1]);

  WorkString = split(lines[LinePointer++], '\t');
  device.MaxSpeed = PApplet.parseInt(WorkString[1]);    

  WorkString = split(lines[LinePointer++], '\t');
  device.MaxFrames = PApplet.parseInt(WorkString[1]);     //Max Frames for sequences that use DataFrameVar[] on device
  device.MaxFramesGlobal = PApplet.parseInt(WorkString[2]);  //max frames for POVs currently and anything that uses GlobalFrameVar
  
  WorkString = split(lines[LinePointer++], '\t');
  device.ConfigFlagsStr = WorkString[1];    

  WorkString = split(lines[LinePointer++], '\t');
  device.ConfigBytesStr = WorkString[1];    

  WorkString = split(lines[LinePointer++], '\t');
  device.DMXModeLabels = new String[16];
  device.DMXModeLabels = split(WorkString[1], ',');

  //println("DMXLables: "+device.DMXModeLabels.length+"    "+device.DMXModeLabels[0]+" "+device.DMXModeLabels[1]);

  device.DMXModeAmt = device.DMXModeLabels.length;     
  
  //Init GUI Elements
  DMXModesDD.labels = device.DMXModeLabels;   //drop down
  DMXModesDD.numStrs = device.DMXModeAmt;   //drop down
  
  //Pixel Type Modes if applicable
  WorkString = split(lines[LinePointer++], '\t');
  device.ListedChipsets = PApplet.parseInt( WorkString[1]);   
  
  if(device.ListedChipsets > 0)
  {  
    String[] tempPixLables = new String[16];

    tempPixLables = new String[device.ListedChipsets]; //use same array as DMX  
    PixelChipsetID = new int[device.ListedChipsets];
    
    //pixel Type Compatibility
    for (int i = LinePointer; i != (LinePointer+device.ListedChipsets);i++)
    {
      WorkString = split(lines[i], '\t');
      tempPixLables[i-LinePointer] = WorkString[1];
      PixelChipsetID[i-LinePointer] = PApplet.parseInt(WorkString[2]);
    }

    PixelChipsetDD.labels = tempPixLables;
    PixelChipsetDD.numStrs = device.ListedChipsets;  
  }  //end if()

  //added v.2c for linked double duty variables
  device.ArrayCastSize = device.Channels; //cSoftwareMaxFrames
  if(device.MaxFrames > device.Channels) device.ArrayCastSize = device.MaxFrames;
      
  // =================   END DATA LOAD  =========================================================
  
  if(device.BasicVersion > 0)
  {
    switch(device.BasicVersion)
    {
    case 1: //4chanDMX
      device.BasicMaxSequences = 8;
      device.BasicSeqLength = (device.BasicMaxSequences*4)+(device.BasicMaxSequences * (device.Channels*2))+device.DataSpace+(device.DataSpace/device.Channels);
      //Misc+(startvals+framevars)+dataspace
      
      println("============"+device.BasicSeqLength+"================");
      
      AutoDetectDD.labels = cDDAutoDetectModesT2;
      AutoDetectDD.numStrs = cDDAutoDetectModesT2.length;
      break;
    case 2: //octo
    case 3: //four chan
      device.BasicMaxSequences = 32;
      //device.BasicSeqLength = (device.BasicMaxSequences*4)+(device.BasicMaxSequences * (device.Channels*2))+device.DataSpace+(device.DataSpace/device.Channels);
      //Misc+(startvals+framevars)+dataspace
      
      AutoDetectDD.labels = cDDAutoDetectModesT3;
      AutoDetectDD.numStrs = cDDAutoDetectModesT3.length;
      break;    
    }
  }//end if
  
  //set Memory/DataSpace label
  if(device.DataSpace > 1000000) device.MemorySpaceString = nf((((float)device.DataSpace/1048576)),1,2)+"MB";
  else device.MemorySpaceString = nf((((float)device.DataSpace/1024)),1,2)+"KB";  
  
  //Prevents UserChannelAmt from being reset when reconnecting
 // if(holdHardwareID != device.HardwareID)  UserChannelAmt = device.Channels;

  println("Device File Successfully Loaded - Starting Configuration Modules");
  
  //================  init Hardware Usage Modules  ============================================
  
  modulePosX = cModueOffsetX; //constants
  modulePosY = cModueOffsetY;
  
  moduleHoldX = 0;
  moduleHoldY = 0;
  moduleHoldSz = 0;
  
  for (int i = 0; i != InputModule.length;i++) InputModule[i].Enabled = false; //reset all modules to disabled
  
  //  device.ConfigFlagsStr = "fDisModeMSBa,fDisModeLSBa,fLEDModeMSBa,fLEDModeLSBa,fAutoDetectID,fAutoDetect,f8or16Bit,fRS485Enable,fMasterEnable,fMasterFullPkts,fMaster16Bit,fAccMeterAutoShutOff,fAccMeterMovement,fAccMeterDoubleTap,fDualModeCom,none";
  //  device.ConfigBytesStr = "bDMXAdrMSB,bDMXAdrLSB,bDMXModeID,bBaudRateID,bGlobalColorShift,bPixelChipset,bAccmeterTimeoutVal,bAccmeterDoubleTapVal";

  ConfigFlags = 0; //reset flags here.....
  DevStrArray = split(device.ConfigFlagsStr, ',');
  String[] ModuleString = new String[16];  //reset
  String[] ByteStrArray = new String[16];
  ByteStrArray = split(device.ConfigBytesStr, ',');  
  //Now single delimenated string is an array, the pointer is the bit number in ConfigFlags  
  
  
  //First handle prioritized tiles ====================================================================================
  
  //z counts from high to low, and checks for tiles that are not yet placed and are above that number
  for(int z = 260; z > 0; z-=10)
  {  
    for(int y = 0; y != InputModule.length;y++)
    {  
      if(InputModule[y].Enabled == false && InputModule[y].Priority >= z)
      {
        ModuleString = split(InputModule[y].ModuleFlagStr, ','); //fill with bit string

        for(int q = 0; q != ModuleString.length; q++)
        {  
          for(int i = 0; i != DevStrArray.length; i++)
          {  
            if(DevStrArray[i].equals(ModuleString[q]) == true) //if string matches
            {
              if(InputModule[y].Enabled == false) 
              {
                //println("Bit tile placed priority "+z);
                //println("Module Priority: "+InputModule[y].Priority);              
                InputModule[y].PlaceTile();
              }
            }
          } //end i for()
        } //end q for()  
        
        //==============================================================
        
        ModuleString = split(InputModule[y].ModuleByteStr, ','); //refill with byte string
        
        for(int q = 0; q != ModuleString.length; q++)
        {  
          for(int i = 0; i != ByteStrArray.length; i++)
          {  
            if(ByteStrArray[i].equals(ModuleString[q]) == true) //if string matches
            {
              if(InputModule[y].Enabled == false) 
              {
                //println("Byte tile placed priority "+z);
                //println(InputModule[y].Priority);              
                InputModule[y].PlaceTile();
              }
            }
          } //end i for()
        } //end q for()          
      }
    } //end y for()
  }  //end z for()
  

  ///Now place all unprioritized not yet enabled tiles
  for(int y = 0; y != InputModule.length;y++)
  {  
    if(InputModule[y].Enabled == false)
    {
      ModuleString = split(InputModule[y].ModuleFlagStr, ',');

      for(int q = 0; q != ModuleString.length; q++)
      {  
        for(int i = 0; i != DevStrArray.length; i++)
        {  
          if(DevStrArray[i].equals(ModuleString[q]) == true) //if string matches
          {
            if(InputModule[y].Enabled == false) 
            {            
              InputModule[y].PlaceTile();
            }
          }
        } //end i for()
      } //end q for()  
      
      //==============================================================
      
      ModuleString = split(InputModule[y].ModuleByteStr, ',');
      
      for(int q = 0; q != ModuleString.length; q++)
      {  
        for(int i = 0; i != ByteStrArray.length; i++)
        {  
          if(ByteStrArray[i].equals(ModuleString[q]) == true) //if string matches
          {
            if(InputModule[y].Enabled == false) 
            {          
              InputModule[y].PlaceTile();
            }
          }
        } //end i for()
      } //end q for()        
      
    }
  } //end y for()
  //Modules placed and prioritized

  // ================   Set additional define flags that require configurations to be loaded ==============

  //if 16-bit data mode or has 16-bit pack function for serial, live control would also be 16-bit capable
  if(device.Bit16Mode == true || IsOnConfigFlags("f8or16Bit"))   device.Bit16LiveMode = true;

  // ================   grey out unavailable GUI elements ==============


  if(device.BasicVersion > 0)
  {
    if(device.BasicVersion == 1) 
    {
   //   SequenceModeDD.Labels = cDDSequenceModeBasic1;
  //    SequenceModeDD.numStrs = cDDSequenceModeBasic1.length;
    }
    //update sequence type options
    if(device.LinkedSupport == true)
    {
   //   SequenceModeDD.Labels = cDDSequenceModeBasic2;
   //   SequenceModeDD.numStrs = cDDSequenceModeBasic2.length;
    }  
  } //end basicVersion if()

  ShowNotification(1);
  println("Configuration Modules Successfully Loaded - end LoadDeviceFile()");
}//end func()

//==================================================================================================================================================  

public boolean IsOnConfigFlags(String passedStr)
{
  for(int i = 0; i != DevStrArray.length; i++)
  {  
    if(DevStrArray[i].equals(passedStr) == true) //if string matches
    {
      println("IsOnConfigFlags() found "+passedStr);
      return true;
    }
  } //end i for()
  return false;
}//end func  

//==================================================================================================================================================

public void bitSet(int pointer)
{
  ConfigFlags |= (1 << pointer); 
  //println("BitSet "+pointer+"    "+ConfigFlags);  
} //end func

//==================================================================================================================================================

//======================================================================================================

public void ManualDeviceDDCallBack()
{
  switch(comManualDeviceSelectDD.selStr)
  {
  case 0:
    comSerialBaudRateDD.selStr = 0; //set to "USB/None"
    comSerialBaudRateDD.status = 0; //show
    break;
  case 1: //pixel mini
    comSerialBaudRateDD.selStr = 1; //set to 9600
    comSerialBaudRateDD.status = 1; //grey
    break;
  case 2:
    comSerialBaudRateDD.selStr = 2; //set to 19200
    comSerialBaudRateDD.status = 1; //grey
    break;
  }
}

//======================================================================================================

public void txtHandleGenericNumber()
{
  if (textFieldPtr.inputMethod > 0)
  {
    if (PApplet.parseInt(textFieldPtr.label) > textFieldPtr.maxValue) textFieldPtr.label = ""+textFieldPtr.maxValue;
    else if (PApplet.parseInt(textFieldPtr.label) < textFieldPtr.minValue)  textFieldPtr.label = ""+textFieldPtr.minValue;
  }
  textFieldPtr.selected = false;
}

//======================================================================================================

public void genericDDCallBack()
{
  //do nothing
}

//======================================================================================================

public void genericHandlerTextField()
{
  textFieldPtr.selected = false;
}

//======================================================================================================

public void genericNumberInputFieldFloats()
{
  println("generatedTextSize() with "+numberInputFieldPtr.selected);

  switch(numberInputFieldPtr.selected)
  {
  case 0:
    break;
  case 1: //decrement
    if (numberInputFieldPtr.value > numberInputFieldPtr.minValue)   numberInputFieldPtr.value-=0.1f;
    break;
  case 2: //increment
    if (numberInputFieldPtr.value < numberInputFieldPtr.maxValue)   numberInputFieldPtr.value+=0.1f;
    break; 
  case 3: //typed in selection
    numberInputFieldPtr.selected = 0; //clear selection

    numberInputFieldPtr.value = PApplet.parseFloat(numberInputFieldPtr.label); //covert string to float
    if (numberInputFieldPtr.value < numberInputFieldPtr.minValue)  numberInputFieldPtr.value = numberInputFieldPtr.minValue;
    else if (numberInputFieldPtr.value > numberInputFieldPtr.maxValue)   numberInputFieldPtr.value = numberInputFieldPtr.maxValue;
    break;
  } //end switch

  numberInputFieldPtr.label = str(numberInputFieldPtr.value);
  println("value: "+numberInputFieldPtr.value);
} //end func()

//======================================================================================================

public void genericNumberInputField()
{
  println("generatedTextSize() with "+numberInputFieldPtr.selected);

  switch(numberInputFieldPtr.selected)
  {
  case 0:
    break;
  case 1: //decrement
    if (numberInputFieldPtr.value > numberInputFieldPtr.minValue)   numberInputFieldPtr.value--;
    break;
  case 2: //increment
    if (numberInputFieldPtr.value < numberInputFieldPtr.maxValue)   numberInputFieldPtr.value++;
    break; 
  case 3:
    numberInputFieldPtr.selected = 0; //clear selection

    numberInputFieldPtr.value = PApplet.parseInt(numberInputFieldPtr.label); //covert string to int
    if (numberInputFieldPtr.value < numberInputFieldPtr.minValue)  numberInputFieldPtr.value = numberInputFieldPtr.minValue;
    else if (numberInputFieldPtr.value > numberInputFieldPtr.maxValue)   numberInputFieldPtr.value = numberInputFieldPtr.maxValue;
    break;
  } //end switch

  numberInputFieldPtr.label = str(numberInputFieldPtr.value);
  println("value: "+numberInputFieldPtr.value);
} //end func()

//======================================================================================================

//=====================================================================================================

class guiNumberInputField
{
  int xpos; 
  int ypos; 
  int size;
  int fieldWidth;

  int minValue;
  int maxValue;

  int inputMethod; //0 = characters only, 1 = numbers only , 2 = numbers only with trailing %, 3 = floats

  String callBack;

  float value;
  String label;
  int status; //0=show, 1=grey out, 2=hide

  int action; //0 = enter, 1 = decrement, 2 = increment

  int selected; //0 = no selection, 1 = decrement, 2 = textfield, 3 = increment
  boolean MouseOverFlag;

  int colTxtBG;
  int colTxtHL;
  int colTxtStroke;
  int colButtonBG;
  int colButtonStroke;

  guiNumberInputField(int ixpos, int iypos, int isize, int ifieldWidth, int iminValue, int imaxValue, int iinputMethod, String icallBack)
  {
    xpos = ixpos;
    ypos = iypos;   
    size = isize;
    fieldWidth = ifieldWidth;
    minValue = iminValue;
    maxValue = imaxValue;
    inputMethod = iinputMethod;
    callBack = icallBack;

    selected = 0;
    MouseOverFlag = false;

    value = minValue;
    label = ""+value;

    colTxtBG = gui.textFieldBG;
    colTxtHL = gui.textFieldHighlight;
    colTxtStroke = gui.buttonColor;

    colButtonBG = gui.buttonColor;
    colButtonStroke = gui.textColor;
  } //end constructor


  //--------------------------------------------------------------------------

  public void handleStatus()
  {
    noStroke();

    if (status == 1) //grey out
    {
      strokeWeight(2);
      stroke(gui.windowBackground, 200); //grey out button
      fill(gui.windowBackground, 200); //grey out button
  //    rect(xpos, ypos, fieldWidth+(size/2)+(size*2), size);  
  
  rect(xpos+size+(size/4), ypos, fieldWidth, size);
      rect(xpos, ypos, size, size, size/4);  //rounded corners
    rect(xpos+(size*1.5f)+fieldWidth, ypos, size, size, size/4);  
  
      noStroke();
    }
  }

  //--------------------------------------------------------------------------

  public void selectField()
  {
    try {
      deselectTextField(); //handle deselecting previous field if applicable
    }
    catch(Exception e) {
    }

    numberInputFieldPtr = this;
    GlobalLabelStore = label;
    label = "";
    NumberInputFieldActive = true;
  }


  //--------------------------------------------------------------------------

  public void display()
  {
    if (status == 2) return; //element is hidden, do not draw
    pushStyle();

    strokeWeight(1);
    stroke(colTxtStroke);

    /*
    if (selected == true) 
     {  
     fill(colHL);
     
     if (status == 0 && MouseOverFlag == true) 
     {
     fill(colHL, 128);
     }
     } else  
     {  
     fill(colBG);
     
     if (status == 0 && MouseOverFlag == true) 
     {
     fill(colBG, 64);
     }
     }
     */

    strokeWeight(2);
    fill(colTxtBG);

    if (selected == 3)
    {
      stroke(colTxtHL);
    } else       
    {
      stroke(gui.buttonColor);
      if (status == 0 && MouseOverFlag == true) 
      {
        stroke(colTxtHL, 128);
        fill(colTxtBG, 192);
      }
    }
    rect(xpos+size+(size/4), ypos, fieldWidth, size);

    //adjust stroke to on higligh and invert text color
    noStroke();
    fill(0);
    textSize((size/1.5f));
    textAlign(CENTER);

    if (inputMethod == 0 || inputMethod == 3) text(label, xpos+(size*1.25f), ypos+((size-(textAscent()+textDescent()))/1.5f), fieldWidth, size);
    else if (inputMethod == 1) text(""+PApplet.parseInt(label), xpos+(size*1.25f), ypos+((size-(textAscent()+textDescent()))/1.5f), fieldWidth, size); 
    else  if (inputMethod == 2)  text(PApplet.parseInt(label)+"%", xpos+(size*1.25f), ypos+((size-(textAscent()+textDescent()))/1.5f), fieldWidth, size);

    strokeWeight(1);
    stroke(colButtonStroke);


    if (selected == 1) fill(colTxtHL);
    else fill(colButtonBG);
    rect(xpos, ypos, size, size, size/4);  //rounded corners

    if (selected == 2) fill(colTxtHL);
    else fill(colButtonBG);
    rect(xpos+(size*1.5f)+fieldWidth, ypos, size, size, size/4);  

    //adjust stroke to on higlight and invert text color
    fill(colButtonStroke);

    textSize((size/1.75f));
    text("-", xpos+(size/2), ypos+(size/1.5f));
    text("+", xpos+(size*2)+fieldWidth, ypos+(size/1.5f));

    handleStatus();

    if (selected !=3) selected = 0; //buttons do not stay highlighted but text field does
    popStyle();
  }


  public boolean over() 
  {
    if (status != 0) return false;

    //check decrement button    
    if (mouseXS >= xpos && mouseXS <= xpos+size && mouseYS >= ypos && mouseYS <= ypos+size)
    {     

      selectField();
            selected = 1; //decrement
      method(callBack);
      return true;
    } 

    //check decrement button    
    if (mouseXS >= xpos+(size*1.5f)+fieldWidth && mouseXS <= xpos+(size*2.5f)+fieldWidth && mouseYS >= ypos && mouseYS <= ypos+size)
    {     

      selectField();
            selected = 2; //increment
      method(callBack);
      return true;
    } 

    //check text field    
    if (mouseXS >= xpos+(size*1.25f) && mouseXS <= xpos+(size*1.25f)+fieldWidth && mouseYS >= ypos && mouseYS <= ypos+size)
    {     
      selectField();
     selected = 3; //text field
      return true;
    } 

    return false;
  } //end 0ver()
} //end class

//=====================================================================================================

class CheckBox
{
  int xpos; // rect xposition  
  int ypos; // rect yposition
  int bWidth;
  int bHeight;
  int colBG;
  int colHL;
  int colOut;  
  boolean selected;
  int status = 0; //0=show, 1=grey out, 2=hide

  boolean MouseOverFlag;

  CheckBox(int ixpos, int iypos, int ibWidth, int ibHeight, int iColBG, int iColHL, int iColOut, boolean iSelected)
  {
    xpos = ixpos;
    ypos = iypos;  
    bWidth = ibWidth;
    bHeight = ibHeight;
    colBG = iColBG;
    colHL = iColHL;
    colOut = iColOut;
    selected = iSelected;

    CheckBoxList.add(this);
    MouseOverFlag = false;
  }

  public void display()
  {
    strokeWeight(1);
    stroke(colOut);
    fill(colBG);

    if (selected==true) 
    {  
      fill(colBG);

      if (status == 0 && MouseOverFlag == true) 
      {
        fill(colBG, 128);
      }
    } else  
    {  
      fill(colBG);

      if (status == 0 && MouseOverFlag == true) 
      {
        fill(colBG, 64);
      }
    }    
    rect(xpos, ypos, bWidth, bHeight);  


    if (selected==true)
    {
      line(xpos, ypos, xpos+bWidth, ypos+bHeight);
      line(xpos+bWidth, ypos, xpos, ypos+bHeight);
    }

    handleStatus();
  }//end display()


  public void handleStatus()
  {
    noStroke();

    if (status == 1) 
    {
      strokeWeight(2);
      stroke(gui.windowBackground, 200); //grey out button
      fill(gui.windowBackground, 200); //grey out button
      rect(xpos, ypos, bWidth, bHeight);  
      noStroke();
    }
  }

  public boolean over() 
  {
    if ( mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos && mouseYS <= ypos+bHeight && status == 0)
    {     
      selected =! selected;  
      return true;
    } else {
      return false;
    }
  }


  public boolean mouseOver() 
  {
    if ( mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos && mouseYS <= ypos+bHeight && status == 0)
    {     
      MouseOverFlag = true;  
      return true;
    } else {
      MouseOverFlag = false;
      return false;
    }
  }
}//end class

//======================================================================================

class DropDown
{
  //use one over() for detecting the button click and if it is open, then use a different over() with the full tab page size
  String[] labels;
  int numStrs;
  int selStr;
  int xpos; // rect xposition  
  int ypos ; // rect yposition
  int bWidth;
  int bHeight;
  int colBG;
  int colHL;
  int colOut;  
  boolean selected;
  int status; //0=show, 1=grey out, 2=hide

  boolean dirMode; //false = down, true = up

String callBack;

  boolean MouseOverFlag;

  //--------------------------------------------------------------------------

  DropDown(String[] iLabels, int inumStrs, int iSelStr, int ixpos, int iypos, int ibWidth, int ibHeight, int iColBG, int iColHL, int iColOut, boolean iDirMode, String icallBack)
  {
    labels = iLabels; 
    numStrs = inumStrs;
    selStr = iSelStr;
    xpos = ixpos;
    ypos = iypos;  
    bWidth = ibWidth;
    bHeight = ibHeight;
    colBG = iColBG;
    colHL = iColHL;
    colOut = iColOut;
    selected = false;
    dirMode = iDirMode;
    status = 0; //normal
    callBack = icallBack;

    DropDownList.add(this);
  }

  //--------------------------------------------------------------------------

  public void handleStatus()
  {
    noStroke();

    if (status == 1) 
    {
      strokeWeight(2);
      stroke(gui.windowBackground, 200); //grey out button
      fill(gui.windowBackground, 200); //grey out button
      rect(xpos, ypos, bWidth, bHeight);  
      noStroke();
    }
  }  

  //--------------------------------------------------------------------------

  public void displayNormal()
  {
    strokeWeight(1);
    stroke(colOut);

    //    if (Selected==true) fill(ColHL);
    //    else fill(ColBG);

    if (selected==true) 
    {  
      fill(colHL);

      if (status == 0 && MouseOverFlag == true) 
      {
        fill(colHL, 128);
      }
    } else  
    {  
      fill(colBG);

      if (status == 0 && MouseOverFlag == true) 
      {
        fill(colBG, 64);
      }
    }

    rect(xpos, ypos, bWidth, bHeight);  
    //adjust stroke to on higlight and invert text color
    textAlign(LEFT);
    fill(colOut);
    textSize((bHeight/1.75f));
    text(labels[selStr], xpos+(bWidth/20), ypos+((bHeight-(textAscent()+textDescent()))/1.5f), bWidth, bHeight);

    fill(255);
    int tempSize = (bWidth/4);
    if (tempSize >  (bHeight/4)) tempSize = (bHeight);

    //   triangle(xpos+(bWidth-(bWidth/4)), ypos+(bHeight/4), xpos+(bWidth-(bWidth/4))+(bHeight/2), ypos+(bHeight/4), xpos+(bWidth-(bWidth/4))+(bHeight/4), ypos+(bHeight/4)+(bHeight/2));
    triangle(xpos+(bWidth-tempSize), ypos+(bHeight/4), xpos+(bWidth-tempSize)+(bHeight/2), ypos+(bHeight/4), xpos+(bWidth-tempSize)+(bHeight/4), ypos+(bHeight/4)+(bHeight/2));
    noStroke();
  }

  //--------------------------------------------------------------------------

  public void selectDD()
  {
    selected =! selected;  //toggle selected state

    if (selected == true)
    {  
      DropDownPointer = this;
      GlobalDDOpen = true;
      DropDownMouseOverID = selStr;
    } else
    {
      GlobalDDOpen = false; //clear in case
    }
  }

  //--------------------------------------------------------------------------

  public void display()
  {
    if (status == 2) return; //do not display if status = 2 = hide

    pushStyle();

    if (selected == false)
    {
      displayNormal();
    } else
    {
      if (dirMode == false)  
      {  
        //down mode
        fill(colBG);
        displayNormal();
        strokeWeight(1);
        stroke(colOut);  

        for (int i = 0; i != numStrs; i++)
        {
          fill(colBG);
          rect(5+xpos, ypos+(bHeight*(i+1)), bWidth-5, bHeight);
          if (i == selStr) fill(colHL);
          else fill(colBG);
          //drop down drawn

          if (DropDownMouseOverID > 0 && DropDownMouseOverID == (i+1) && MouseOverFlag == true)
          {
            if (i == selStr) fill(colHL, 128);
            else fill(0, 0, 0, 64);//fill(ColBG,64);
          }

          rect(5+xpos, ypos+(bHeight*(i+1)), bWidth-5, bHeight);

          fill(colOut);
          textSize((bHeight/1.75f));
          text(labels[i], 5+xpos+((bWidth - textWidth(labels[i]))/2), ypos+((bHeight-(textAscent()+textDescent()))/1.5f)+(bHeight*(i+1)), bWidth, bHeight);
        }   //end for 
        noStroke();
      } else
      {            
        //Menu upwards
        displayNormal(); //display the button part
        strokeWeight(1);
        stroke(colOut);  

        for (int i = 0; i != numStrs; i++)
        {
          fill(colBG);
          rect(5+xpos, ypos+(bHeight*(i-numStrs)), bWidth-5, bHeight); //draw background first
          if (((numStrs-1)-i) == selStr) fill(colHL);
          else fill(colBG);
          //drop down drawn

          if (DropDownMouseOverID > 0 && DropDownMouseOverID == (i+1) && MouseOverFlag == true)
          {
            if (((numStrs-1)-i) == selStr) fill(colHL, 128);
            else fill(0, 0, 0, 64);//fill(ColBG,64);
          }
          rect(5+xpos, ypos+(bHeight*(i-numStrs)), bWidth-5, bHeight);
        } //end for

        for (int i = 0; i != numStrs; i++)
        {
          fill(colOut);
          textSize((bHeight/1.75f));
          text(labels[i], 5+xpos+((bWidth - textWidth(labels[i]))/2), ypos+((bHeight-(textAscent()+textDescent()))/1.5f)-(bHeight*(i+1)), bWidth, bHeight);
        }   //end for 
        noStroke();
      }
    }

    handleStatus();
    popStyle();
  } //end display()

  //--------------------------------------------------------------------------

  public boolean overOpen()
  {
    if (dirMode == false)  
    {    
      //Down Mode
      if (selected == false)
      {
        return false;
      } else
      {
        for (int i = 0; i != numStrs; i++)
        {
          if (mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos+(bHeight*(i+1)) && mouseYS <= ypos+bHeight+(bHeight*(i+1)) && status == 0)
          {    
            selected = false;
            selStr = i;  
            GlobalDDOpen = false;        
            return true;
          }
        }
        selected = false;
        return false;
      }
    } else
    {  
      if (selected == false)
      {
        return false;
      } else
      {
        for (int i = 0; i != numStrs; i++)
        {
          if (mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos-(bHeight*(i+1)) && mouseYS <= ypos+bHeight-(bHeight*(i-1)) && status == 0)
          {    
            selected = false;
            selStr = i;  
            GlobalDDOpen = false;        
            return true;
          }
        }
        selected = false;
        return false;
      }
    }
  } //end overNew()

  //--------------------------------------------------------------------------

  public boolean over() ///over the icon to open
  {
    if ( mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos && mouseYS <= ypos+bHeight  && status == 0) 
    {    
      if (GlobalDDOpen == true) return false; //if drop down already open, don't let it open another

      selectDD();
      return true;
    } else {
      return false;
    }
  }

  //--------------------------------------------------------------------------

  public boolean mouseOver() ///over the icon to open
  {
    if ( mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos && mouseYS <= ypos+bHeight  && status == 0)   
    {      
      MouseOverFlag = true;  
      return true;
    } else {
      MouseOverFlag = false;
      return false;
    }
  }

  //--------------------------------------------------------------------------

  public boolean mouseOverOpen()
  {
    if (dirMode == false)  
    {    
      //Down Mode
      if (selected == false)
      {
        return false;
      } else
      {
        for (int i = 0; i != numStrs; i++)
        {
          if (mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos+(bHeight*(i+1)) && mouseYS <= ypos+bHeight+(bHeight*(i+1)) && status == 0)
          {    
            MouseOverFlag = true;
            DropDownMouseOverID = i+1;        
            return true;
          }
        }
        DropDownMouseOverID = 0;
        MouseOverFlag = false;
        return false;
      }
    } else
    {  
      if (selected == false)
      {
        return false;
      } else
      {
        for (int i = 0; i != numStrs; i++)
        {
          if (mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos-(bHeight*(i+1)) && mouseYS <= ypos+bHeight-(bHeight*(i-1)) && status == 0)
          {    
            MouseOverFlag = true;
            DropDownMouseOverID = (numStrs-i);        
            return true;
          }
        }
        DropDownMouseOverID = 0;
        MouseOverFlag = false;
        return false;
      }
    }
  } //end overNew()
} //end dropdown class

//======================================================================================

class Button
{
  String Label;
  int xpos; // rect xposition  
  int ypos ; // rect yposition
  int bWidth;
  int bHeight;
  int colBG;
  int colHL;
  int colOut;  
  boolean toggle;
  boolean selected;
  int status; //0=show, 1=grey out, 2=hide

  boolean MouseOverFlag;

  //--------------------------------------------------------------------------

  Button(String iLabel, int ixpos, int iypos, int ibWidth, int ibHeight, int iColBG, int iColHL, int iColOut, boolean iToggle, boolean iSelected, boolean iAllowMouseOver)
  {
    Label = iLabel; 
    xpos = ixpos;
    ypos = iypos;  
    bWidth = ibWidth;
    bHeight = ibHeight;
    colBG = iColBG;
    colHL = iColHL;
    colOut = iColOut;
    toggle = iToggle;
    selected = iSelected;
    status = 0; //init at 0

    //Setup mouse over list
    if (iAllowMouseOver == true) ButtonList.add(this);
    MouseOverFlag = false;
  }

  //--------------------------------------------------------------------------

  public void handleStatus()
  {
    noStroke();

    if (status == 1) 
    {
      strokeWeight(2);
      stroke(gui.windowBackground, 200); //grey out button
      fill(gui.windowBackground, 200); //grey out button
      rect(xpos, ypos, bWidth, bHeight);  
      noStroke();
    }
  }

  //--------------------------------------------------------------------------

  public void displayOutline()
  {
    strokeWeight(3);
    stroke(colOut);

    fill(255, 255, 255, 0); //transparent fill
    rect(xpos, ypos, bWidth, bHeight, bHeight/4);
  }

  //--------------------------------------------------------------------------

  public void display()
  {
    if (status == 2) return; //element is hidden, do not draw
    pushStyle();
    strokeWeight(1);
    stroke(colOut);

    if (selected==true) 
    {  
      fill(colHL);

      if (status == 0 && MouseOverFlag == true) 
      {
        fill(colHL, 128);
      }
    } else  
    {  
      fill(colBG);

      if (status == 0 && MouseOverFlag == true) 
      {
        fill(colBG, 64);
      }
    }

    textAlign(LEFT);
    rect(xpos, ypos, bWidth, bHeight, bHeight/4);  
    //adjust stroke to on higlight and invert text color
    fill(colOut);
    textSize((bHeight/1.75f));
    text(Label, xpos+((bWidth - textWidth(Label))/2), ypos+((bHeight-(textAscent()+textDescent()))/1.5f), bWidth, bHeight);

    handleStatus();

    if (toggle==false) selected = false;
    popStyle();
  }

  //--------------------------------------------------------------------------

  public boolean over() 
  {
    if (mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos && mouseYS <= ypos+bHeight && status == 0) 
    {      
      selected =! selected;  
      return true;
    } else {
      return false;
    }
  } //end over()

  //--------------------------------------------------------------------------

  public boolean mouseOver() //does not toggle selected
  {
    if (mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos && mouseYS <= ypos+bHeight && status == 0) 
    {      
      MouseOverFlag = true;  
      return true;
    } else {
      MouseOverFlag = false;
      return false;
    }
  } //end over()
} //end class

//======================================================================================

class SliderBar
{
  int  SetNumber;
  int xpos; // rect xposition  
  int ypos ; // rect yposition
  int slWidth; // single bar slWidthth
  int slHeight; // rect height
  int Value; 
  int Minimum; 
  int Maximum;
  int BgColor;
  int slHandleColor;
  int slSelectColor; 
  int slStrokeColor;
  boolean CircleHandle;  //ture circle, false rectangle
  boolean HorizVertToggle;
  boolean ScrollBarToggle;
  boolean GraphicDisplay;

  String methodStr;

  int status;

  //gotta move the int variables into float variables
  float temp1=slWidth;
  float temp2=Maximum;
  float temp3=Value;
  float temp4=Minimum;
  float MathVar=0;

  SliderBar(int iSetNumber, int ixp, int iyp, int iw, int ih, int iValue, int iMin, int iMax, int iBgColor, int islHandleColor, int islSelectColor, int islStrokeColor, boolean iCircleHandle, boolean iHorizVertToggle, boolean iScrollBarToggle, boolean iGraphicDisplay, String imethodStr) 
  {   
    xpos = ixp;
    ypos = iyp;  
    slWidth = iw;
    slHeight = ih;
    Value=iValue;
    Maximum=iMax;
    Minimum=iMin;
    BgColor=iBgColor;
    slHandleColor=islHandleColor;
    slSelectColor=islSelectColor;
    slStrokeColor=islStrokeColor;
    CircleHandle=iCircleHandle;      //false is rectangle, true is circle
    HorizVertToggle=iHorizVertToggle; //false is horiz, true is vert
    ScrollBarToggle=iScrollBarToggle; //false is normal, true is scrollbar style
    GraphicDisplay = iGraphicDisplay;
    methodStr =imethodStr;
    SetNumber = iSetNumber;

    status = 0;  //0=show, 1=grey out, 2=hide
  }

  public void setValue(int passedVal)
  {
    if (passedVal > Maximum) { //check so it doesn't drag off the end
      passedVal = Maximum;
    }
    if (passedVal < Minimum) { //check so it doesn't drag off the end
      passedVal = Minimum;
    }
    Value = passedVal;
  }//end setValue() 


  public int getUnsignedChar()
  {
    //returns 8 bit unsigned number 0 - 255, based on float 
    return PApplet.parseInt(255*getFloat());
  }


  public float getFloat()
  {
    return ((float)Value / (float)Maximum);
  }


  public int getValue()
  {
    if (Value > Maximum) { //check so it doesn't drag off the end
      Value = Maximum;
    }
    if (Value < Minimum) { //check so it doesn't drag off the end
      Value = Minimum;
    }  

    return Value;
  }

  public void runMethod()
  {
    move(mouseXS, mouseYS);
    try {
      method(methodStr); //run function that is called everytime the slider is adjusted
    }
    catch(Exception e) { 
      println("Slider Method not found");
    }
  }



  public void move(int uX, int uY) 
  {
    //uX = round((float)uX / SF);
    //uY = round((float)uY / SF);    
    if (ScrollBarToggle == true)
    {
      //scrollbar mode
      if (HorizVertToggle == false) //horizontal
      {
        //this function handles position of the slider's handle for Verticle
        temp1 = slWidth - (slHeight * 2);   
        temp3 = Value;

        MathVar = ((((uX-xpos)-slHeight) / temp1)-1) *-1; //creates a scale factor based on mouse position 
        MathVar = Maximum - (MathVar*(Maximum-Minimum));   

        if (MathVar > Maximum) { //check so it doesn't drag off the end
          MathVar=Maximum;
        }
        if (MathVar < Minimum) { //check so it doesn't drag off the end
          MathVar=Minimum;
        }    

        Value=PApplet.parseInt(MathVar); //set the value based on mouse movement
      }//end if
      else
      {
        //this function handles position of the slider's handle for Verticle
        temp1 = slHeight - (slWidth * 2);   
        temp3 = Value;

        MathVar = ((((uY-ypos)-slWidth) / temp1)-1) *-1; //creates a scale factor based on mouse position  
        MathVar = Maximum - (MathVar*(Maximum-Minimum));   

        if (MathVar > Maximum) { //check so it doesn't drag off the end
          MathVar=Maximum;
        }
        if (MathVar < Minimum) { //check so it doesn't drag off the end
          MathVar=Minimum;
        }  
        Value=PApplet.parseInt(MathVar); //set the value based on mouse movement
      }
    } else
    {
      //regular usage
      if (HorizVertToggle == false) //horizontal
      {
        //this function handles position of the slider's handle
        temp1 = slWidth;   
        temp3 = Value;
        temp4 = Minimum;

        MathVar = (((uX-xpos) / temp1)-1) *-1; //creates a scale factor based on mouse position 
        //relative to slider position
        MathVar = Maximum - (MathVar*(Maximum-Minimum));   

        if (MathVar > Maximum) { //check so it doesn't drag off the end
          MathVar=Maximum;
        }

        if (MathVar < Minimum) { //check so it doesn't drag off the end
          MathVar=Minimum;
        }  

        Value = PApplet.parseInt(MathVar); //set the value based on mouse movement
      }//end if
      else
      {
        println("NO SLIDER MOVE() HANDLER");
      }
    }
  } //end move()


  public void display() 
  {   
    if (status == 2) return; //element is hidden, do not draw

    pushStyle();

    if (ScrollBarToggle == true)//ScrollBarToggle places buttons at either end, if false it has no end space
    {
      if (HorizVertToggle == false) //horizontal
      { 
        stroke(slStrokeColor);
        strokeWeight(1);

        //circle handles are displayed differently so there are if/else statements to set it correctly
        fill(BgColor); //covers up the old slider with a blank rectangle, BgColor set in declaration
        rect(xpos, ypos, slWidth, slHeight);   //main rect

        fill(200); 
        rect(xpos+(slHeight/10)+slHeight, ypos+(slHeight/10), slWidth-(slHeight/5)-(slHeight*2), (slHeight-(slHeight/5)));   //middle    
        rect(xpos+(slHeight/10), ypos+(slHeight/10), slHeight-(slHeight/5), slHeight-(slHeight/5)); //left rect  
        rect(xpos+(slHeight/10)+(slWidth-slHeight), ypos+(slHeight/10), slHeight-(slHeight/5), slHeight-(slHeight/5));  //right rect 

        stroke(slStrokeColor);
        fill(slHandleColor);
        strokeWeight(1);
        //top
        triangle(xpos+slHeight-(slHeight/5), ypos+(slHeight/5), xpos+slHeight-(slHeight/5), ypos+(slHeight-(slHeight/5)), xpos+(slHeight/5), ypos+(slHeight/2));  
        //right
        triangle(xpos+slWidth-slHeight+(slHeight/5), ypos+(slHeight/5), xpos+slWidth-slHeight+(slHeight/5), ypos+(slHeight-(slHeight/5)), xpos+slWidth-slHeight+slHeight-(slHeight/5), ypos+(slHeight/2));
        //Handle

        temp1 = (slWidth - (slHeight*3));
        temp2 = Maximum;  //move variables into float
        temp3 = Value;    //or do they don't compute correctly
        temp4 = Minimum;      
        float MathVar;

        MathVar = temp1 / (temp2 - temp4) * (temp3-temp4) + slHeight/2+slHeight;

        fill(slHandleColor); //places the slider handle
        ellipse(xpos+MathVar, ypos+(slHeight/2), slHeight-2, slHeight-2); //Circle instead of rectangle
      } else
      {
        stroke(slStrokeColor);
        strokeWeight(1);

        //circle handles are displayed differently so there are if/else statements to set it correctly
        fill(BgColor); //covers up the old slider with a blank rectangle, BgColor set in declaration
        rect(xpos, ypos, slWidth, slHeight);  

        fill(200); 
        rect(xpos+(slWidth/10), ypos+(slWidth/10)+slWidth, slWidth-(slWidth/5), (slHeight-(slWidth/5))-(slWidth*2));   
        rect(xpos+(slWidth/10), ypos+(slWidth/10), slWidth-(slWidth/5), slWidth-(slWidth/5));   
        rect(xpos+(slWidth/10), ypos+(slWidth/10)+slHeight-slWidth, slWidth-(slWidth/5), slWidth-(slWidth/5));   

        stroke(slStrokeColor);
        fill(slHandleColor);
        strokeWeight(1);
        //top
        triangle(xpos+(slWidth/5), ypos+(slWidth/4)+(slWidth/2), xpos+(slWidth/2), ypos+(slWidth/5), (xpos+slWidth) - (slWidth/5), ypos+(slWidth/4)+(slWidth/2));
        //bottom
        triangle(xpos+(slWidth/5), ypos+(slWidth/4)+slHeight-slWidth, xpos+(slWidth/2), ypos+slHeight-(slWidth/5), (xpos+slWidth) - (slWidth/5), ypos+(slWidth/4)+slHeight-slWidth);

        temp1 = (slHeight - (slWidth*3));
        temp2 = Maximum;  //move variables into float
        temp3 = Value;    //or do they don't compute correctly
        temp4 = Minimum;

        //println(temp1+" : "+temp2+" : "+temp3+" : "+temp4);

        //Handle
        float MathVar;
        MathVar = temp1 / (temp2 - temp4) * (temp3-temp4) + slWidth/2+slWidth;
        fill(slHandleColor); //places the slider handle
        ellipse(xpos+(slWidth/2), ypos+MathVar, slWidth-2, slWidth-2); //Circle instead of rectangle

        // rect(xpos, ypos+MathVar, slWidth, slWidth); //and comment this line for circle
        noStroke();
      }
    } else
    {
      stroke(slStrokeColor);
      strokeWeight(1);

      //circle handles are displayed differently so there are if/else statements to set it correctly
      fill(BgColor); //covers up the old slider with a blank rectangle, BgColor set in declaration
      rect(xpos, ypos, slWidth, slHeight);   

      if (CircleHandle==true) temp1 = slWidth - slHeight;
      else temp1=slWidth-(slWidth/20);

      temp2=Maximum;  //move variables into float
      temp3=Value;    //or do they don't compute correctly
      temp4 = Minimum;

      float MathVar;

      if (CircleHandle==true)
        MathVar = temp1 / (temp2 - temp4) * (temp3-temp4) + slHeight/2;
      else  
      MathVar = temp1 / (temp2 - temp4) * (temp3-temp4); 

      fill(color(slSelectColor)); //draws the Selcolor over the selected area
      rect(xpos, ypos, MathVar, slHeight);

      fill(slHandleColor); //places the slider handle
      if (CircleHandle==true)   ellipse(xpos+MathVar, ypos+(slHeight/2), slHeight-1, slHeight-1); //Circle instead of rectangle
      else   rect((xpos+MathVar), ypos, slWidth/20, slHeight); //and comment this line for circle
    }
    popStyle();
  }//end display()

  public void selectSlider()
  {
    SliderPointer = this;
    GlobalDragging = true;
  }//end selectSlider

  public boolean over() 
  {
    if (mouseXS >= xpos && mouseXS <= xpos+slWidth && mouseYS >= ypos && mouseYS <= ypos+slHeight) 
    {          
      return true;
    } else {
      return false;
    }
  }
}

//=====================================================================================

class TextField
{
  String label;
  int xpos; // rect xposition  
  int ypos ; // rect yposition
  int bWidth;
  int bHeight;
  int colBG;
  int colHL;

  int inputMethod; //0 = characters only, 1 = numbers only , 2 = numbers only with trailing %, 3 = floats
  int minValue; //either min number, or minim characters
  int maxValue; //either max number, or maximum characters
  boolean toggle;
  boolean selected;
  String callBack;

  int status; //0=show, 1=grey out, 2=hide

  boolean MouseOverFlag;

  TextField(String iLabel, int ixpos, int iypos, int ibWidth, int ibHeight, int iColBG, int iColHL, int iInputMethod, int iminValue, int imaxValue, boolean iToggle, boolean iSelected, String itxtfCallback)
  {
    label = iLabel; 
    xpos = ixpos;
    ypos = iypos;  
    bWidth = ibWidth;
    bHeight = ibHeight;
    colBG = iColBG;
    colHL = iColHL;
    toggle = iToggle;
    selected = iSelected;
    inputMethod = iInputMethod;
    minValue = iminValue; 
    maxValue = imaxValue; 
    callBack = itxtfCallback;
    status = 0; //init at show

    //if(sEnableMouseOver == 1) 
    TextFieldList.add(this);
    MouseOverFlag = false;
  }

  public void handleStatus()
  {
    noStroke();

    if (status == 1) 
    {
      strokeWeight(2);
      stroke(gui.windowBackground, 200); //grey out button
      fill(gui.windowBackground, 200); //grey out button
      rect(xpos, ypos, bWidth, bHeight);  
      noStroke();
    }
  }  

  public void display()
  {
    if (status == 2) return; //element is hidden, do not draw

    //translate(0, 0);
    strokeWeight(2);
    fill(colBG);

    if (selected==true)
    {
      stroke(colHL);
    } else       
    {
      stroke(gui.buttonColor);
      if (status == 0 && MouseOverFlag == true) 
      {
        stroke(colHL, 128);
        fill(colBG, 192);
      }
    }
    rect(xpos, ypos, bWidth, bHeight);

    //adjust stroke to on higligh and invert text color
    noStroke();
    fill(0);
    textSize((bHeight/1.75f));

    textAlign(LEFT);
    if (inputMethod != 2) text(label, xpos+((bWidth - textWidth(label))/2), ypos+((bHeight-(textAscent()+textDescent()))/1.5f), bWidth, bHeight);
    else  text(label+"%", xpos+((bWidth - textWidth(label+1))/2), ypos+((bHeight-(textAscent()+textDescent()))/1.5f), bWidth, bHeight);

    handleStatus();

    if (toggle==false) selected = false;
  }


  public void SelectField()
  {
    try {
      //textFieldPtr.selected = false; //was commented out
      deselectTextField(); //handle deselecting previous field if applicable
    }
    catch(Exception e) {
    }

    textFieldPtr = this;
    GlobalLabelStore = label;
    label = "";
    selected = true; 

    TextFieldActive = true;
  }


  public boolean over() 
  {
    if (mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos && mouseYS <= ypos+bHeight && status == 0) 
    {     
      selected =! selected;    
      return true;
    } else {
      return false;
    }
  }

  public boolean mouseOver() //does not toggle selected
  {
    if (mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos && mouseYS <= ypos+bHeight && status == 0) 
    {     
      MouseOverFlag = true;    
      return true;
    } else {
      MouseOverFlag = false;
      return false;
    }
  } //end over()
} 

//============================================================================================

//===================================================================================================  

public void keyPressed()
{
  
  //------------------------------------------------------------------------------------------------
  
  if (NumberInputFieldActive == true)
  {

    if (key == ENTER || key == RETURN)
    {
      NumberInputFieldActive = false;
      method(numberInputFieldPtr.callBack);
      return;
    } //end enter/return if
    else if (key == BACKSPACE)
    {
      try {
        println("Trying");
        numberInputFieldPtr.label = numberInputFieldPtr.label.substring(0, numberInputFieldPtr.label.length()-1);
      }
      catch(Exception e) {
      }  
      return;
    }   


    if (numberInputFieldPtr.inputMethod == 0)
    {
      if (numberInputFieldPtr.label.length() > numberInputFieldPtr.maxValue) return;
      //Makes sure its a valid character otherwise it will be stored(but not seen) in .label as well
      if ((byte)key > 0 && (byte)key < 127) numberInputFieldPtr.label =  ""+numberInputFieldPtr.label + PApplet.parseChar(key);
      else println("KEY IGNORED"); //DEBUG
      return;
    } else
    {
      if (key=='1' || key== '2' || key=='3' || key=='4' || key=='5' || key=='6' || key=='7' || key=='8' || key=='9' || key=='0' || key=='-' || (key =='.' && numberInputFieldPtr.inputMethod == 3))
      {
        numberInputFieldPtr.label =  ""+numberInputFieldPtr.label + PApplet.parseChar(key);
        return;
      }
    }
    return;
  }  //end NumberInputFieldActive if()

  //------------------------------------------------------------------------------------------------


  if (TextFieldActive == true)
  {
    if (key == ENTER || key == RETURN)
    {
      TextFieldActive = false;
      method(textFieldPtr.callBack);
      return;
    } //end enter/return if
    else if (key == BACKSPACE)
    {
      try {
        println("Trying");
        textFieldPtr.label = textFieldPtr.label.substring(0, textFieldPtr.label.length()-1);
      }
      catch(Exception e) {
      }  
      return;
    }  


    if (textFieldPtr.inputMethod == 0)
    {
      if (textFieldPtr.label.length() > textFieldPtr.maxValue) return;
      //Makes sure its a valid character otherwise it will be stored(but not seen) in .label as well
      if ((byte)key > 0 && (byte)key < 127) textFieldPtr.label =  ""+textFieldPtr.label + PApplet.parseChar(key);
      else println("KEY IGNORED"); //DEBUG
      return;
    } else
    {
      if (key=='1' || key== '2' || key=='3' || key=='4' || key=='5' || key=='6' || key=='7' || key=='8' || key=='9' || key=='0' || key=='-' || (key =='.' && textFieldPtr.inputMethod == 3))
      {
        textFieldPtr.label =  ""+textFieldPtr.label + PApplet.parseChar(key);
        return;
      }
    }


    if (key == ENTER || key == RETURN)
    {
      TextFieldActive = false;
      method(textFieldPtr.callBack);
    } //end enter/return if
    return;
  }  //end NumberInputFieldActive if()


  //------------------------------------------------------------------------------------------------

 // if (key == 'x')  PrintDebugMessages = !PrintDebugMessages;



  if (key == 'i') 
  {
  }


  if (key == ESC) 
  {
    key = 0; //have to alter it or if it gets to the end of the function with key=escape it closes the application
   // CloseOverLayMenu();
    return; //double up preventing escape for closing application
  }

  if (key == CODED) 
  {
    switch(keyCode)
    {
    case UP:
      if(DropDownPointer.selStr > 0) DropDownPointer.selStr--;
      break;

    case DOWN:
      if(DropDownPointer.selStr < (DropDownPointer.numStrs-1) ) DropDownPointer.selStr++;
      break;
    case RIGHT:

      break;

    case LEFT:

      break;
    }
  }
}//end keyPressed()

//===================================================================================================  

class SpecHWModules
{
  int IDnumber;
  String Label;
  int xpos;
  int ypos;
  int bWidth;
  int bHeight;

  int Priority;

  String ModuleFlagStr;
  String ModuleByteStr;

  boolean Enabled;

  SpecHWModules(int iIDnumber, String iLabel, int ixpos, int iypos, int ibWidth, int ibHeight, String iModuleFlagStr, String iModuleByteStr, int iPriority)
  {
    IDnumber = iIDnumber;
    Label = iLabel;
    xpos = ixpos;
    ypos = iypos;  
    bWidth = ibWidth;
    bHeight = ibHeight;
    ModuleFlagStr = iModuleFlagStr;
    ModuleByteStr = iModuleByteStr;
    Priority = iPriority;
    Enabled = false;
  }

  public void display()
  {
    if (Enabled == false) return;

    stroke(0);
    strokeWeight(2);
    fill(gui.windowStroke, 200);
    rect(xpos, ypos, bWidth, bHeight);
    noStroke();

    switch(IDnumber)
    {
    case 0:  //device.RS485Capable
      HardwareButtons[0].display();    
      break;    
    case 1:    //LED indicator 2
      LEDIndicatorModeBDD.display();
      if (LEDIndicatorModeBDD.selected == true)  DropDownPointer = LEDIndicatorModeBDD;    
      break;      
    case 2:  //device.GlobalColorShift    
      HardwareColSwapDD.display();  
      if (HardwareColSwapDD.selected == true)  DropDownPointer = HardwareColSwapDD;  
      break;      
    case 3:  //Pixel Type
      PixelChipsetDD.display();  
      if (PixelChipsetDD.selected == true)  DropDownPointer = PixelChipsetDD;
      break;        
    case 4:  //device.HardwareID
      HardwareButtons[2].display();  
      break;      
    case 5: //DMX Adress
      TextFields[2].display(); //DMX Address
      break;      
    case 6:
      Label = "DMX Reception\nMode - ID Number: "+DMXModesDD.selStr;
      DMXModesDD.display(); //has to be last so it overlays everything else
      break;  
    case 7:
      MasterModeDD.display();
      break;
    case 8:
      HardwareCheckBox[0].display();
      fill(0);
      textAlign(LEFT);
      text("16-Bit Serial", HardwareCheckBox[0].xpos+25, HardwareCheckBox[0].ypos+15);    
      Label = "Serial Baud Rate:\nID Number: "+BaudRateDD.selStr;
      BaudRateDD.display();  
      break;
    case 9:
      LEDIndicatorModeADD.display();
      break;  
    case 10:
      DisplayModeDD.display();
      break;  
    case 11:
      AutoDetectDD.display();  
      break;      
    case 12: //fGammaCorrectionEnabled
      HardwareButtons[3].display();
      break;      
    case 13: //Automatic Shut Off/Resume Time Out Value
      AccMeterAutoDetectDD.display();  
      AccMeterAutoDetectModeDD.display();  
      break;        
    case 14: //Enable Accelerometer Speed Adjustment
      HardwareButtons[4].display();
      break;
    case 15: //Enable Double Tap Button Press
      AccMeterTapDD.display();  
      break;
    case 16: //Power Down Time Out
      PowerDownTimeOutDD.display();  
      break;    
    case 17: //Packet Clone

      break;  
    case 18: //Fade Transition

      break;
    case 19: //Dual Communication Mode
      HardwareCheckBox[1].display();  
      fill(0);
      textAlign(LEFT);
      text("Enable", HardwareCheckBox[1].xpos-45, HardwareCheckBox[1].ypos+15); 
      break;  
    case 20: //PWM Freq A
      PWMFreqADD.display();
      break;
    case 21: //Stand-Alone A
      StandAloneADD.display();
      TextFields[13].display();
      fill(0);
      textAlign(LEFT);
      textSize(12);
      text("Enter Sequence ID#:", TextFields[13].xpos-90, TextFields[13].ypos-5);
      break;
    case 22: //PWM Profile
      PWMProfileADD.display();
      break;  
    case 23: //I.R. Remote addon card
      IRRemoteModeDD.display();
      break;      
      //start added modules
    case 24: //DMX Pixel Amount:
      TextFields[4].display();
      break;
    case 25: //"Enable Activity LED:
      HardwareCheckBox[2].display();
      break;
    case 26: //DMX Decoder Mode:
      DMXDecoderModeDD.display();
      break;
    case 27: //Enable Serial Color Swap:
      HardwareCheckBox[3].display();
      break;
    case 28: //Select Color Order For Chipset:
      HardwareColSwapMicroDD.display();
      break;
    case 29: //End-Of-Frame Timer
      TextFields[3].display();
      break;
    case 30: //mini Serial Baud Rate
      BaudRateMiniV4DD.display();
      break;
    case 31: //DMX Timeout auto release
      DMXAutoReleaseDD.display();
      break;   
     case 32: //Legancy Signal Loss Auto Detection
      AutoReleaseLegacyDD.display();
      break;   
     case 33: //Legacy Reception mode. DMX or serial
      ReceptionModeLegacyDD.display();
      break;  
    }

    textAlign(CENTER);
    textSize(12);  
    textLeading(14);
    fill(0);
    text(Label, xpos+5, ypos+10, bWidth-10, bHeight-10);
  } //end display

  public void InitPositions()
  {

    switch(IDnumber)
    {
    case 0:  //device.RS485Capable
      HardwareButtons[0].xpos = xpos+30;
      HardwareButtons[0].ypos = ypos+55;
      break;    
    case 1:  //second LED indicator
      LEDIndicatorModeBDD.xpos = xpos+10;
      LEDIndicatorModeBDD.ypos = (ypos+(bHeight))-40;
      break;      
    case 2:  //device.GlobalColorShift  
      HardwareColSwapDD.xpos = xpos+10;            
      HardwareColSwapDD.ypos = ypos+60;      
      break;      
    case 3:  //Pixel Type
      PixelChipsetDD.xpos = xpos+10;    
      PixelChipsetDD.ypos = ypos+25;    
      break;        
    case 4:  //device.HardwareID
      HardwareButtons[2].xpos = xpos+60;
      HardwareButtons[2].ypos = ypos+55;  
      break;  
    case 5:  
      TextFields[2].xpos = xpos+20;
      TextFields[2].ypos = (ypos+(bHeight))-50;        
      break;        
    case 6: //DMX Reception Mode  
      DMXModesDD.xpos = xpos+5;
      DMXModesDD.ypos = (ypos+(bHeight))-35;      
      break;      
    case 7: //DMX Master Mode
      if (device.Bit16Mode == true)
      {
        MasterModeDD.labels = cDDMasterModesMultiBit;
        MasterModeDD.numStrs = cDDMasterModesMultiBit.length;
      } else
      {
        MasterModeDD.labels = cDDMasterModes8Bit;
        MasterModeDD.numStrs = cDDMasterModes8Bit.length;
      }
      MasterModeDD.xpos = xpos+10;
      MasterModeDD.ypos = (ypos+(bHeight))-35;  
      break;      
    case 8:
      BaudRateDD.xpos = xpos+10;
      BaudRateDD.ypos = (ypos+(bHeight))-52;  

      HardwareCheckBox[0].xpos = xpos+10;
      HardwareCheckBox[0].ypos = (ypos+(bHeight))-25;  

      //set checkbox grey out or not if enabled
      if (IsOnConfigFlags("f8or16Bit") == true) HardwareCheckBox[0].status = 0; //0=show, 1=grey out, 2=hide
      else HardwareCheckBox[0].status = 1; //0=show, 1=grey out, 2=hide    
      break;  
    case 9:
      LEDIndicatorModeADD.xpos = xpos+10;
      LEDIndicatorModeADD.ypos = (ypos+(bHeight))-40;
      break;  
    case 10:
      DisplayModeDD.xpos = xpos+10;
      DisplayModeDD.ypos = (ypos+(bHeight))-40;
      break;  
    case 11:
      AutoDetectDD.xpos = xpos+10;
      AutoDetectDD.ypos = (ypos+(bHeight))-40;      
      break;      
    case 12://fGammaCorrectionEnabled
      HardwareButtons[3].xpos = xpos+30;
      HardwareButtons[3].ypos = ypos+55;  
      break;  
    case 13: //Automatic Shut Off/Resume Time Out Value
      AccMeterAutoDetectDD.xpos = xpos+20;    
      AccMeterAutoDetectDD.ypos = ypos+65;    
      AccMeterAutoDetectModeDD.xpos = xpos+20;    
      AccMeterAutoDetectModeDD.ypos = ypos+40;
      break;        
    case 14: //Enable Accelerometer Speed Adjustment
      HardwareButtons[4].xpos = xpos+30;
      HardwareButtons[4].ypos = ypos+45;  
      break;
    case 15: //Enable Double Tap Button Press
      AccMeterTapDD.xpos = xpos+20;    
      AccMeterTapDD.ypos = ypos+55;  
      break;    
    case 16: //Power Down Time Out  
      PowerDownTimeOutDD.xpos = xpos+10;
      PowerDownTimeOutDD.ypos = (ypos+(bHeight))-40;  
      break;
    case 17: //empty

      break;  
    case 18: //empty

      break;  
    case 19: //Dual Communication Mode
      HardwareCheckBox[1].xpos=xpos+70;  
      HardwareCheckBox[1].ypos=ypos+65;  
      break;  
    case 20: //PWM Freq A
      PWMFreqADD.xpos=xpos+10;
      PWMFreqADD.ypos=(ypos+(bHeight))-40;  
      break;
    case 21: //Stand-Alone A
      StandAloneADD.xpos=xpos+5;  
      StandAloneADD.ypos=(ypos+(bHeight))-35;      
      TextFields[13].xpos=xpos+125;  
      TextFields[13].ypos=(ypos+(bHeight))-35;  
      break;
    case 22: //PWM Profile
      PWMProfileADD.xpos = xpos+10;
      PWMProfileADD.ypos = (ypos+(bHeight))-35;      
      break;      
    case 23: //I.R. Remote Add on
      IRRemoteModeDD.xpos = xpos+10;
      IRRemoteModeDD.ypos = (ypos+(bHeight))-35;      
      break;
    case 24: //DMX Pixel Amount:
      TextFields[4].xpos = xpos+20;
      TextFields[4].ypos = (ypos+(bHeight))-50;   
      break;
    case 25: //"Enable Activity LED:
      HardwareCheckBox[2].xpos=xpos+50;  
      HardwareCheckBox[2].ypos=ypos+40;  
      break;
    case 26: //DMX Decoder Mode:
      DMXDecoderModeDD.xpos = xpos+10;
      DMXDecoderModeDD.ypos = (ypos+(bHeight))-35;         
      break;
    case 27: //Enable Serial Color Swap:
      HardwareCheckBox[3].xpos=xpos+50;  
      HardwareCheckBox[3].ypos=ypos+50;  
      break;
    case 28: //Select Color Order For Chipset:
      HardwareColSwapMicroDD.xpos = xpos+10;
      HardwareColSwapMicroDD.ypos = (ypos+(bHeight))-35;      
      break;
    case 29: //End-Of-Frame Timer
      TextFields[3].xpos = xpos+20;
      TextFields[3].ypos = (ypos+(bHeight))-35;   
      break;
    case 30: //mini Serial Baud Rate
      BaudRateMiniV4DD.xpos = xpos+10;
      BaudRateMiniV4DD.ypos = (ypos+(bHeight))-45;         
      break;
     case 31: //DMX Timeout auto release
      DMXAutoReleaseDD.xpos = xpos+10;
      DMXAutoReleaseDD.ypos = (ypos+(bHeight))-35;        
      break;     
     case 32: //Legancy Signal Loss Auto Detection
      AutoReleaseLegacyDD.xpos = xpos+10;
      AutoReleaseLegacyDD.ypos = (ypos+(bHeight))-35;        
      break;   
     case 33: //Legacy Reception mode. DMX or serial
      ReceptionModeLegacyDD.xpos = xpos+10;
      ReceptionModeLegacyDD.ypos = (ypos+(bHeight))-35;        
      break;     
    }
  }//end InitPositions


  public int GetBitValues(int passedID)
  {

    switch(IDnumber)
    {
    case 0:  //device.RS485Capable
      //"fRS485Enable,f8or16Bit"
      if (HardwareButtons[0].selected && passedID == 0)  return 1;//fRS485Enable
      if (HardwareCheckBox[0].selected && passedID == 1) return 1;//f8or16FBit
      break;    
    case 1: //LED Indicator Mode 2

      break;      
    case 2:  //bGlobalColorShift

      break;      
    case 3:  //Pixel Chipset

      break;        
    case 4: //Enable Serial Reception Instead of DMX

      break;  
    case 5:  //DMX Adress

      break;        
    case 6: //DMX Reception

      break;      
    case 7: //DMX Master Mode

      if (IsOnConfigFlags("fMaster16Bit") == true) 
      {
        //"fMasterEnable,fMasterFullPkts,fMaster16Bit"
        switch(MasterModeDD.selStr)
        {
        case 0: //None
          //both ID flag and Autodetect stay 0

          return 0;
        case 1: //Master Full Pkts 8-Bit
          //temp2Flags = temp2Flags | 0x01; //fMasterEnable
          //temp2Flags = temp2Flags | 0x02; //fMasterFullPkts(1 is Full, 0 is partial)
          if (passedID == 0)       return 1;//fMasterEnable
          if (passedID == 1)       return 1;//fMasterFullPkts(1 is Full, 0 is partial)
          if (passedID == 2)       return 0;//fMaster16Bit  stays 0            
          //fMaster16Bit stays 0  
          break;
        case 2: //Master Full Pkts 16-Bit
          //temp2Flags = temp2Flags | 0x01; //fMasterEnable
          //temp2Flags = temp2Flags | 0x02; //fMasterFullPkts(1 is Full, 0 is partial)  
          //temp2Flags = temp2Flags | 0x04; //fMaster16Bit
          if (passedID == 0)       return 1;//fMasterEnable
          if (passedID == 1)       return 1;//fMasterFullPkts(1 is Full, 0 is partial)
          if (passedID == 2)       return 1;//fMaster16Bit  stays 0            

          break;
        case 3:  //Master Partial Pkts 8-Bit
          //temp2Flags = temp2Flags | 0x01; //fMasterEnable  
          if (passedID == 0)       return 1;//fMasterEnable
          if (passedID == 1)       return 0;//fMasterFullPkts(1 is Full, 0 is partial) 
          if (passedID == 2)       return 0;//fMaster16Bit  stays 0      
          break;  
        case 4:  //Master Partial Pkts 16-Bit
          //temp2Flags = temp2Flags | 0x01; //fMasterEnable  
          //temp2Flags = temp2Flags | 0x04; //    
          if (passedID == 0)       return 1;//fMasterEnable
          if (passedID == 1)       return 0;//fMasterFullPkts(1 is Full, 0 is partial)
          if (passedID == 2)       return 1;//fMaster16Bit  
          break;
        }//end switch
      } else
      {
        //"fMasterEnable,fMasterFullPkts"
        switch(MasterModeDD.selStr)
        {
        case 0: //None
          //both ID flags  
          return 0;
        case 1: //Master Full Packets
          //temp2Flags = temp2Flags | 0x01; //fMasterEnable
          //temp2Flags = temp2Flags | 0x02; //fMasterFullPkts(1 is Full, 0 is partial)
          if (passedID == 0)       return 1;//fMasterEnable
          if (passedID == 1)       return 1; //fMasterFullPkts

          break;
        case 2: //Master Partial Packets
          //temp2Flags = temp2Flags | 0x01; //fMasterEnable        
          if (passedID == 0)       return 1;//fMasterEnable
          if (passedID == 1)       return 0;//fMasterFullPkts(1 is Full, 0 is partial) - leave it 0
          break;
        }//end switch
      }

      break;      
    case 8: //Serial Baud Rate
      if (HardwareCheckBox[0].selected == true) return 1;  
      return 0;
      //break;  
    case 9: //LED Inidicator A
      //11 "Both", 01"Stand-Alone", 10"RX", 00"Disabled"
      //"fLEDModeMSBa,fLEDModeLSBa"
      switch(LEDIndicatorModeADD.selStr) //LEDMode
      {
      case 0: //Both
        //tempFlags = tempFlags | 0x0C;
        if (passedID == 0)       return 1;
        if (passedID == 1)       return 1;
        break;
      case 1: //Stand-Alone
        //tempFlags = tempFlags | 0x04;
        if (passedID == 0)       return 0;
        if (passedID == 1)       return 1;      
        break;
      case 2: //RX only
        //tempFlags = tempFlags | 0x08;
        if (passedID == 0)       return 1;
        if (passedID == 1)       return 0;        
        break;
      case 3: //disabled
        //no change both 00
        return 0;
      }
      break;  
    case 10: //External LED Display Mode
      //"fDisModeMSBa,fDisModeLSBa"
      switch(DisplayModeDD.selStr)
      {
      case 0: //Always On
        //00 is always on
        return 0;
      case 1: //Countdown
        if (passedID == 0)       return 1;
        if (passedID == 1)       return 0;
        break;
      case 2: //Dim
        if (passedID == 0)       return 0;
        if (passedID == 1)       return 1;
        break;
      case 3: //both
        if (passedID == 0)       return 1;
        if (passedID == 1)       return 1;
        break;
      }
      break;  
    case 11: //Standard Auto-Detect
      switch(AutoDetectDD.selStr)
      {
      case 0: //None
        //both ID flag and Autodetect stay 0
        return 0;
      case 1: //Autodetect DMX
        //tempFlags = tempFlags | 0x10; //Set AutoDetectID Flag to 1
        //tempFlags = tempFlags | 0x20; //Set AutoDetectEnable Flag to 1
        if (passedID == 0)       return 1;
        if (passedID == 1)       return 1;
        break;
      case 2: //Autodetect Serial
        //AutoDetectID Flag stays 0 for serial
        //tempFlags = tempFlags | 0x20; //Set AutoDetectEnable Flag to 1
        if (passedID == 0)       return 0;
        if (passedID == 1)       return 1;
        break;
      }//end switch
      break;      
    case 12: //Gamma Correction Enabled
      if (HardwareButtons[3].selected == true) return 1;
      return 0;
    case 13: //Automatic Shut Off/Resume Time Out Value
      if (passedID == 0) if (AccMeterAutoDetectDD.selStr > 0) return 1;
      if (passedID == 1) if (AccMeterAutoDetectModeDD.selStr == 1) return 1;
      //if(AccMeterAutoDetectDD.selStr > 0)  return 1; //fAccMeterAutoShutOff
      //if(AccMeterAutoDetectModeDD.selStr > 0)  return 1; //fAccMeterAutoShutOffMode
      return 0; //break;        
    case 14: //Enable Accelerometer Speed Adjustment
      if (HardwareButtons[4].selected)  return 1;//fAccMeterMovement
      return 0; //break;
    case 15: //Enable Double Tap Button Press
      if (AccMeterTapDD.selStr > 0) return 1;//fAccMeterDoubleTap
      return 0; //break;    
    case 16: //Power Down Time Out  

      break;
    case 17:

      break;  
    case 18:

      break;  
    case 19: //Dual Communication Mode
      if (HardwareCheckBox[1].selected == true) return 1; //fDualModeCom
      return 0; //break;  
    case 20: //PWM Freq A

      break;
    case 21: //Stand-Alone A

      break;
    case 22: //PWM Profile


      break;    
    case 23: //IR Remote Addon
      switch(IRRemoteModeDD.selStr)
      {
      case 0: //None
        //both ID flags  0
        return 0;
      case 1: //Enabled, 19200, 1-0
        if (passedID == 0)   return 1;
        if (passedID == 1)   return 0;  
        break;
      case 2: //Enabled, 250k, 1-1  
        if (passedID == 0)   return 1;
        if (passedID == 1)   return 1;
        break;
      }//end switch
      break;

      //start added modules
    case 24: //DMX Pixel Amount:

      break;
    case 25: //"Enable Activity LED:
      if (HardwareCheckBox[2].selected == true) return 1; //fActivityLEDEnable
      else return 0;
      //break;
    case 26: //DMX Decoder Mode:
      switch(DMXDecoderModeDD.selStr)
      {
      case 0: //Disabled
        if (passedID == 0)   return 0;
        if (passedID == 1)   return 0;  
        break;
      case 1: //Full Packets
        if (passedID == 0)   return 1;  //fDMXDecoderModeFull
        if (passedID == 1)   return 0; //fDMXDecoderModeEnable
        break;    
      case 2: //Force Mode
        if (passedID == 0)   return 0; //fDMXDecoderModeFull
        if (passedID == 1)   return 1; //fDMXDecoderModeEnable
        break;
      case 3: //Force Mode, Full Packets
        if (passedID == 0)   return 1; //fDMXDecoderModeFull
        if (passedID == 1)   return 1; //fDMXDecoderModeEnable  
        break;
      }
      break;
    case 27: //Enable Serial Color Swap:
      if (HardwareCheckBox[3].selected == true) return 1; //fEnableSerialColorSwap
      break;
    case 28: //Select Color Order For Chipset:
      switch(HardwareColSwapMicroDD.selStr)
      {
      case 0: //none/RGB
        if (passedID == 0)   return 0;
        if (passedID == 1)   return 0;  
        if (passedID == 2)   return 0;
        if (passedID == 3)   return 0;          
        break;
      case 1: //G>R>B
        if (passedID == 0)   return 1;
        if (passedID == 1)   return 0;  
        if (passedID == 2)   return 0;
        if (passedID == 3)   return 0;         
        break;       
      case 2: //G>R>B>W"
        if (passedID == 0)   return 0;
        if (passedID == 1)   return 1;  
        if (passedID == 2)   return 0;
        if (passedID == 3)   return 0; 
        break;
      case 3: //"B>G>R"
        if (passedID == 0)   return 0;
        if (passedID == 1)   return 0;  
        if (passedID == 2)   return 1;
        if (passedID == 3)   return 0; 
        break;
      case 4: // "B>R>G"
        if (passedID == 0)   return 0;
        if (passedID == 1)   return 0;  
        if (passedID == 2)   return 0;
        if (passedID == 3)   return 1; 
        break;
      }//end switch
      break;
    case 29: //End-Of-Frame Timer

      break;
    case 30: //mini Serial Baud Rate

      break;
  case 31: //DMX Timeout auto release
    switch(DMXAutoReleaseDD.selStr)
    {
    case 0: //None
      //both ID flags  0
      return 0;
    case 1: //Enabled, Restore Sequence
      if(passedID == 0)   return 1;
      if(passedID == 1)   return 0;  
      break;
    case 2: //Enabled, Play Idle  
      if(passedID == 0)   return 1;
      if(passedID == 1)   return 1;
      break;
    }//end switch
    break;
   case 32: //Legancy Signal Loss Auto Detection
    if(AutoReleaseLegacyDD.selStr == 0) return 0;
    else return 1;
    //break;   
   case 33: //Legacy Reception mode. DMX or serial
    if(ReceptionModeLegacyDD.selStr == 0) return 0;
    else return 1;    
    //break;  
    }  //end switch 

    println("Did not find InputModule[] to GetBitValues() from with ID "+IDnumber);
    return 0;
  } //end GetBitValues()


  public int GetByteValues(int passedID)
  {

    switch(IDnumber)
    {
    case 0:  //device.RS485Capable

      break;    
    case 1:

      break;      
    case 2:  //bGlobalColorShift
      return (HardwareColSwapDD.selStr); //Global Color Swap    
    case 3:  //Pixel Chipset
      return (PixelChipsetID[PixelChipsetDD.selStr]); //Pixel Type    
    case 4: //Enable Serial Reception Instead of DMX

      break;  
    case 5:  //DMX Adress
      if (passedID == 0)   return ((PApplet.parseInt(TextFields[2].label) >> 8) & 0xFF); //DMXAdr MSB
      if (passedID == 1)   return (PApplet.parseInt(TextFields[2].label) & 0xFF); //DMXAdr LSB
      break;        
    case 6: //DMX Reception Mode
      return DMXModesDD.selStr;  //DMX Mode ID    
    case 7: //DMX Master Mode

      break;      
    case 8: //Serial Baud Rate
      return BaudRateDD.selStr; //BaudRateID
    case 9: //LED Inidicator A

      break;  
    case 10: //External LED Display Mode

      break;  
    case 11: //Standard Auto-Detect

      break;      
    case 12: //Gamma Correction Enabled

      break;  
    case 13: //Automatic Shut Off/Resume Time Out Value
      return (cAccMeterShutOffVals[AccMeterAutoDetectDD.selStr]);   //Timeoutval      
    case 14: //Enable Accelerometer Speed Adjustment

      break;
    case 15: //Enable Double Tap Button Press

      return (AccMeterTapDD.selStr);  //Double Tap Mode  
    case 16: //Power Down Time Out  

      break;
    case 17:

      break;  
    case 18:

      break;  
    case 19: //Dual Communication Mode

      break; 
    case 20: //PWM Freq LegacyA
      return (PWMFreqADD.selStr+4); //4,5,6,7 are loaded into T6CON
    case 21: //Stand-Alone Legacy A
      switch(StandAloneADD.selStr) //either 10-42 for sequences, 1 for hold, 2 for blank, 3 blank with fade out
      {
      case 0: //hold
      case 1: //blank
      case 2: //fade out
        return (StandAloneADD.selStr+1); //0 is null so add 1
      case 3: //stand alone
        return ((PApplet.parseInt(TextFields[13].label)+9)); //10+ is sequences, base 0, textfield should be at least 1
      }  //end switch
      break;
    case 22: //PWM Profile
      return (PWMProfileADD.selStr);  //16-bit lamp PWM Profile  
    case 23: //IR Remote Add On

      break;
      //start added modules
    case 24: //DMX Pixel Amount:
    
    //CHECK IF IN RIGHT ORDER, THIS CONTROLLER HAS LSB FIRST!
      if (passedID == 0)   return ((PApplet.parseInt(TextFields[4].label) >> 8) & 0xFF); //DMXAdr MSB
      if (passedID == 1)   return (PApplet.parseInt(TextFields[4].label) & 0xFF); //DMXAdr LSB
      break;
    case 25: //Enable Activity LED:

      break;
    case 26: //DMX Decoder Mode:

      break;
    case 27: //Enable Serial Color Swap:

      break;
    case 28: //Select Color Order For Chipset:

      break;
    case 29: //End-Of-Frame Timer
      return PApplet.parseInt((PApplet.parseInt(TextFields[3].label) * 1000)  / 128);
      //break;
    case 30: //mini Serial Baud Rate
      return BaudRateMiniV4DD.selStr; //BaudRateID
  case 31: //DMX Timeout auto release
    
    break;
   case 32: //Legancy Signal Loss Auto Detection

    break;   
   case 33: //Legacy Reception mode. DMX or serial

    break;  
    }  //end switch  

    println("Did not find InputModule[] to GetByteValues() from with ID "+IDnumber);
    return 0;
  } //end GetByteValues()


  public boolean over() 
  {
    if (Enabled == false) return false;  

    if (mouseXS >= xpos && mouseXS <= xpos+bWidth && mouseYS >= ypos && mouseYS <= ypos+bHeight) 
    {      
      return true;
    } else {
      return false;
    }
  }

  public void PlaceTile()
  {
    if (InputModule[IDnumber].bWidth < moduleHoldSz)
    {
      //println("Module "+IDnumber+" with width of: "+InputModule[IDnumber].bWidth+" can fit at Y: "+moduleHoldY+", X: "+moduleHoldX+" with available: "+moduleHoldSz);  

      InputModule[IDnumber].xpos = moduleHoldX;    
      InputModule[IDnumber].ypos = moduleHoldY;

      moduleHoldX = 0;
      moduleHoldY = 0;
      moduleHoldSz = 0;
    } else if ((modulePosX + InputModule[IDnumber].bWidth) < 800)  
    {
      //println("IDnumber: "+IDnumber+" : "+modulePosX);

      InputModule[IDnumber].xpos = modulePosX;    
      modulePosX += (InputModule[IDnumber].bWidth+10);
      InputModule[IDnumber].ypos = modulePosY;
    } else 
    {
      //saves previous row extra space and will try to fit something in
      if ((800 - modulePosX) > moduleHoldSz)
      {
        //if new row size is larger than last, use it, otherwise utilize largest space  
        moduleHoldX = modulePosX;
        moduleHoldY = modulePosY;
        moduleHoldSz = 800 - modulePosX;    
        //println("remaining Module Size: "+moduleHoldSz);
      }

      modulePosY += 100;
      InputModule[IDnumber].xpos = cModueOffsetX;
      modulePosX = cModueOffsetX + (InputModule[IDnumber].bWidth+10);
      InputModule[IDnumber].ypos = modulePosY;
    }

    InputModule[IDnumber].Enabled = true;
    InputModule[IDnumber].InitPositions();
  } //end func
}  //end class

//======================================================================================================================

public void deselectTextField()
{
  if (TextFieldActive == true)
  {
    if (textFieldPtr.selected == true) 
    {
      textFieldPtr.selected = false; 
      textFieldPtr.label = GlobalLabelStore; //restore
    }
    TextFieldActive = false;
  }

  //---------------------------------------------------------

  if (NumberInputFieldActive == true)
  {
    if (numberInputFieldPtr.selected > 0) 
    {
      numberInputFieldPtr.selected = 0; 
      numberInputFieldPtr.label = GlobalLabelStore; //restore
    }

    NumberInputFieldActive = false;
  }
} //end mousePressedHandleTextFields

//======================================================================================================================

public void mousePressed() 
{
  mouseXS = round((float)mouseX / SF);
  mouseYS = round((float)mouseY / SF);

  if (mouseButton == RIGHT) 

  {
    println("X: "+mouseX+"   Y: "+mouseY);

    //---------------------------------------------------------

    //---------------------------------------------------------
  } //end right click
  else     
  {
    //regular left click
    //---------------------------------------------------------

    if (ConfirmBoxIDNum > 0)
    {

      if (ConfirmButtonYes.over())
      {
        switch(ConfirmBoxIDNum)
        {
        case 0: //null
          break;
        case 1: //Upload Configurations      
          if(device.HardwareID == 60 || device.HardwareID == 61)   SendConfigurationsNonAurora(); //pixel micro and pixel mini
          else if (device.BasicVersion == 2) RequestConfigurationUploadType2(); //octosequencer
            else if (device.BasicVersion == 3) RequestConfigurationUploadLegacy();
          else RequestConfigurationUpload(); //send configurations normally for all other devices.
          
          ShowNotification(4);
          break;
        } //end switch

        ConfirmBoxIDNum = 0; //close
      } //end confirm yes

      if (ConfirmButtonNo.over()) ConfirmBoxIDNum = 0; //close

      return;
    } //end confirm box if

    //---------------------------------------------------------

    //If a drop down is open handle first, to prevent click-throughs to other elements
    if (GlobalDDOpen == true)   
    {
      //handle before anything
      if (DropDownPointer.overOpen())
      {
        println("Pressed Open DropDown "+DropDownPointer.selStr);
        method(DropDownPointer.callBack);
      } else
      {
        //else a click was detected that wasn't on the open drop down, close it and leave
        DropDownPointer.selected = false;  
        GlobalDDOpen = false;
      }
      return; //whether it was a click on the open drop down, or somewhere not on the drop down, leave now, no other clicks should be detected
    }

    //---------------------------------------------------------

    if (OverlayMenuID > 0)
    {
      if (guiOverlayMenus[OverlayMenuID].mousePressed()) return; //no need to check over()
      //  if (menuCloseButton.over())  CloseOverLayMenu(); //close the open menu

      deselectTextField(); //handle deselecting previous field if applicable();

      if (guiOverlayMenus[OverlayMenuID].forceOverride == 0) return;
    }

    //---------------------------------------------------------

    if (comSerialPortDD.over()) return;

    if (devConnected == false) 
    {
      if (comManualDeviceSelectDD.over()) return;
      if (comSerialBaudRateDD.over()) return;
    }

    if (comAutoScanButton.over()) return;

    if (comConnectButton.over()) 
    {
      //setup baud rate based on user selections

      if (comSerialBaudRateDD.selStr == 0) ProgramBaudRate = 0;//USB/None selected, use any baud rate, doesn't matter
      else 
      {
        //A baud rate is selected
        ProgramBaudRate = PApplet.parseInt(comSerialBaudRateDD.labels[comSerialBaudRateDD.selStr]);
      }

      if (devConnected == false) OpenCOMPort(serialPort.list()[comSerialPortDD.selStr]); 
      else CloseCOMPort();
      return;
    }

    if (comUploadConfigurations.over())
    {
      SetConfirmBox(1);
    }

    if (CheckModuleElements() == true) return;
  } //end left click else
} //end mousePressed

//======================================================================================================================

public void mouseReleased()
{
  //Update scaled mouse everytime before the values are used
  mouseXS = round((float)mouseX / SF);
  mouseYS = round((float)mouseY / SF);

  GlobalDragging = false; //incase a slider is being dragged
} //end mouseReleased

//======================================================================================================================

public void mouseDragged()
{
  //The mouseDragged() function is called once every time the mouse moves while a mouse button is pressed. (If a button is not being pressed, mouseMoved() is called instead.) 
  //println("MouseDragging");
  //  if(GlobalDragging == true) redraw();
}

//======================================================================================================================

public void mouseMoved()
{
  mouseXS = round((float)mouseX / SF);
  mouseYS = round((float)mouseY / SF);

  //---------------------------------------------------------

  if (software.mouseOverEnabled == true)
  {
    //Mouse Over Functions
    // Fairly ineffiecent, it will detect any object, even those not shown, so it will always redraw on mouse moved

    for (int i = 0; i != DropDownList.size(); i++) 
    { 
      if (DropDownList.get(i).mouseOver()) return; //no other way to highlight, have to redraw
      if (DropDownList.get(i).mouseOverOpen()) return; //no other way to highlight, have to redraw
    } //end for

    //---------------------------------------------------------

    if (GlobalDDOpen == true)   return; //if a drop down is open, don't bother mousing over any other elements, just leave

    //---------------------------------------------------------

    for (int i = 0; i != ButtonList.size(); i++) 
    { 
      // An ArrayList doesn't know what it is storing so we have to cast the object coming out
      PointerButton = ButtonList.get(i);  
      if (PointerButton.mouseOver()) return;
    } //end for

    //---------------------------------------------------------
  } //end sEnableMouseOver if()
} //end mouseMoved

//======================================================================================================================

public void mouseWheel()
{
  println("mouseWheel()");
}

//============================================================================================

public class DeviceObj
{
  int HardwareID; //unique hardware ID number
  String Name;  //Stores the readable name for the controller as string
  int HardwareVersion;
  int FirmwareVersion;
  int FirmwareRevision;
  int BootloaderVersion;

  //Sequence defines
  int Channels; //maximum amount of channels
  int MaxSpeed;  
  int MaxIndexedSequences;
  int IndexMemoryModel;
  int MaxFrames;
  int MaxFramesGlobal;
  
  //Memory defines
  int DataSpace;
  int HWPVSpace;
  int HPVBlockSize;
  int SeqMemoryBlockSize; 
  int IndexBlockSize; 
  int CurrentBlockSize;
  int EraseBlockSize;

  boolean LockChannelAmt; //disallow per sequence data size, only device.channels
  int GammaCorrection;
  boolean LinkedSupport;  
  boolean Bit16Mode;  //flag to set 16-bit data mode
  boolean FillUpperData;
  boolean Bit16Processor;
  boolean Bit16LiveMode;
  
  int BasicVersion;
  int BasicMaxSequences; 
  int BasicSeqLength;
  int AccelerometerMode;

  int ArrayCastSize; //can't always use device channels, because of double duty for linked variables
  
//How are these loaded differently???????????????  
  int DMXModeAmt;
  String[] DMXModeLabels;
  int ListedChipsets;

  //Not loaded from Device Files
  int UserConfiguredIDNum; //set by user saved in firwmare  
  int UserIdleSequence;  //Held is save file, sent with index
  
  //scratch
  String ConfigFlagsStr;
  String ConfigBytesStr;
  
  String WebpageURL;
  String MemorySpaceString; //stores byte to kb/mb conversion string

  DeviceObj(String iName)
  {
    HardwareID = 0;
    Name = iName;
    HardwareVersion = 0;
    FirmwareVersion = 0;
    FirmwareRevision = 0;  
    Channels = 0;
    DataSpace = 0;
    BootloaderVersion = 0;
    UserConfiguredIDNum = 0;
    UserIdleSequence = 0;

    MaxFrames = 0;
    MaxFramesGlobal = 0;
  
    MaxIndexedSequences = 8; //default, will get set when device file is loaded
    
    EraseBlockSize = 0;
    HWPVSpace = 0;
    DMXModeAmt = 0;
    GammaCorrection = 0;
    LinkedSupport = false;
    Bit16Mode = false;
    DMXModeLabels = new String[1];   
  }
} //end object

//============================================================================

class SoftwareObj
{
  int frameRate;

  int GUIWidth;
  int GUIHeight;

  boolean mouseOverEnabled;

  //--------------------------------------------------------------------------

  SoftwareObj()
  {
    mouseOverEnabled = false;
  }
  //--------------------------------------------------------------------------
} //end object

//=================================================================================

public class GUIObj
{
  int windowBackground;
    int windowStroke;
  int layerBackground;
  int textColor;
  int textMenuColor;

  int buttonColor;
  int buttonHighlightColor;
  int menuBackground;

  int textFieldHighlight;
  int textFieldBG;
  
 
  //--------------------------------------------------------------------------

  GUIObj()
  {
    windowBackground = color(65, 65, 65);
    windowStroke  = color(255);
    layerBackground = color(100, 100, 100);
    textColor = color(255);
    buttonColor = color(0, 0, 100);
    buttonHighlightColor = color(100, 0, 100);
    
    menuBackground  = color(200);
    textMenuColor = color(0);
    
    textFieldBG = color(255); //white
    textFieldHighlight = color(200, 40, 40);
  }
} //end gui class

//=======================================================================================================================  

class OverlayMenu
{
  int ObjID;
  int xpos;
  int ypos;
  int Height;
  int Width;

  int autoPosition; //0 = use xpos & ypos, 1 = center X & Y, 2 = center Y alight right, 3 = center Y align left
  int forceOverride; //0 = only menu elements can be clicked, 1 = any exposed elements can be clicked

  int xOffset, yOffset;

  //constructor
  OverlayMenu(int iObjID, int ixpos, int iypos, int ibWidth, int ibHeight)
  {
    ObjID = iObjID;
    xpos = ixpos;
    ypos = iypos;  
    Width = ibWidth;
    Height = ibHeight;

    autoPosition = 1; //set as default
    forceOverride = 0;
  } //end constructor

  //-----------------------------------------------------------------------------------------------

  public void display()
  {
    pushStyle();
    fill(gui.menuBackground);
    stroke(0);
    strokeWeight(4);


    //draws background of menu
    switch(autoPosition) 
    {
    case 0:
    default:
      rect(xpos, ypos, Width, Height, 20);
      break;
    case 1:
      rect((width-Width)/2, (height-Height)/2, Width, Height, 20);
      break;
    }//end switch

    //========================================

    switch(ObjID)
    {
    case 1: 

      break;
    } //end ObjID switch

    menuCloseButton.xpos = xOffset+(Width-100);
    menuCloseButton.ypos = yOffset+20;
    menuCloseButton.display();

    popStyle();
  } //end display()

  //-----------------------------------------------------------------------------------------------

  public void initMenu()
  {

    //=======================
    switch(autoPosition) 
    {
    case 0:
    default:
      xOffset = xpos;
      yOffset = ypos;
      break;
    case 1: //= center X&Y
      xOffset = (width-Width)/2;
      yOffset = (height-Height)/2;
      break;
    }//end switch   
    //====================

    switch(ObjID)
    {
    case 1:
    
       break;
    } //end ObjID switch
  } //end display()

  //-----------------------------------------------------------------------------------------------

  public boolean over() 
  {
    if (mouseXS >= xpos && mouseXS <= xpos+Width && mouseYS >= ypos && mouseYS <= ypos+Height) 
    {      
      return true;
    } else {
      return false;
    }
  } //end over()

  //-----------------------------------------------------------------------------------------------

  public boolean overEllipse() 
  {
    if (mouseXS >= xpos-(Width/2) && mouseXS <= xpos+(Width/2) && mouseYS >= ypos-(Height/2) && mouseYS <= ypos+(Height/2)) 
    {      
      return true;
    } 
    else {
      return false;
    }
  } //end over()  

  //-----------------------------------------------------------------------------------------------

  public boolean mousePressed()
  {

    switch(ObjID)
    {
    case 1:
      break;  
    } //end switch
    return false;
  } //end mouse pressed
} //end overlayMenu class
  public void settings() {  size(800, 600); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "NLEDConfigurationSoftwarev2b" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
